# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba
import numba.typing.arraydecl

import awkward1.layout
from ..._numba import cpu, util, content

@numba.extending.typeof_impl.register(awkward1.layout.IndexedArray32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedArrayU32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedArray64)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedOptionArray32)
@numba.extending.typeof_impl.register(awkward1.layout.IndexedOptionArray64)
def typeof(val, c):
    import awkward1._numba.types
    return IndexedArrayType(numba.typeof(numpy.asarray(val.index)), numba.typeof(val.content), val.isoption, numba.typeof(val.identities), util.dict2parameters(val.parameters))

class IndexedArrayType(content.ContentType):
    def __init__(self, indextpe, contenttpe, isoption, identitiestpe, parameters):
        assert isinstance(parameters, tuple)
        super(IndexedArrayType, self).__init__(name="ak::Indexed{0}Array{1}{2}Type({3}, identities={4}, parameters={5})".format("Option" if isoption else "", "" if indextpe.dtype.signed else "U", indextpe.dtype.bitwidth, contenttpe.name, identitiestpe.name, util.parameters2str(parameters)))
        self.indextpe = indextpe
        self.contenttpe = contenttpe
        self.isoption = isoption
        self.identitiestpe = identitiestpe
        self.parameters = parameters

    @property
    def bitwidth(self):
        return self.indextpe.dtype.bitwidth

    @property
    def indexname(self):
        return ("" if self.indextpe.dtype.signed else "U") + str(self.indextpe.dtype.bitwidth)

    @property
    def ndim(self):
        return self.contenttpe.ndim

    def getitem_int(self):
        if self.isoption:
            return numba.types.optional(self.contenttpe.getitem_int())
        else:
            return self.contenttpe.getitem_int()

    def getitem_range(self):
        return self

    def getitem_str(self, key):
        return IndexedArrayType(self.indextpe, self.contenttpe.getitem_str(key), self.isoption, self.identitiestpe, ())

    def getitem_tuple(self, wheretpe):
        import awkward1._numba.array.listarray
        nexttpe = awkward1._numba.array.listarray.ListArrayType(util.index64tpe, util.index64tpe, self, numba.none, ())
        out = nexttpe.getitem_next(wheretpe, False)
        return out.getitem_int()

    def getitem_next(self, wheretpe, isadvanced):
        import awkward1._numba.array.regulararray
        if len(wheretpe.types) == 0:
            return self
        headtpe = wheretpe.types[0]
        tailtpe = numba.types.Tuple(wheretpe.types[1:])

        if isinstance(headtpe, (numba.types.Integer, numba.types.SliceType, numba.types.Array)):
            if self.isoption:
                contenttpe = self.contenttpe.carry().getitem_next(wheretpe, isadvanced)
                return IndexedArrayType(self.indextpe, contenttpe, self.isoption, self.identitiestpe, self.parameters)
            else:
                return self.contenttpe.carry().getitem_next(wheretpe, isadvanced)

        elif isinstance(headtpe, numba.types.StringLiteral):
            return self.getitem_str(headtpe.literal_value).getitem_next(tailtpe, isadvanced)

        elif isinstance(headtpe, numba.types.EllipsisType):
            raise NotImplementedError("ellipsis")

        elif isinstance(headtpe, type(numba.typeof(numpy.newaxis))):
            raise NotImplementedError("newaxis")

        else:
            raise AssertionError(headtpe)

    def carry(self):
        return self

    @property
    def lower_len(self):
        return lower_len

    @property
    def lower_getitem_nothing(self):
        return content.lower_getitem_nothing

    @property
    def lower_getitem_int(self):
        return lower_getitem_int

    @property
    def lower_getitem_range(self):
        return lower_getitem_range

    @property
    def lower_getitem_str(self):
        return lower_getitem_str

    @property
    def lower_getitem_next(self):
        return lower_getitem_next

    @property
    def lower_carry(self):
        return lower_carry

@numba.extending.register_model(IndexedArrayType)
class IndexedArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("index", fe_type.indextpe),
                   ("content", fe_type.contenttpe)]
        if fe_type.identitiestpe != numba.none:
            members.append(("identities", fe_type.identitiestpe))
        super(IndexedArrayModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(IndexedArrayType)
def unbox(tpe, obj, c):
    asarray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(numpy.asarray))
    index_obj = c.pyapi.object_getattr_string(obj, "index")
    content_obj = c.pyapi.object_getattr_string(obj, "content")
    indexarray_obj = c.pyapi.call_function_objargs(asarray_obj, (index_obj,))
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    proxyout.index = c.pyapi.to_native_value(tpe.indextpe, indexarray_obj).value
    proxyout.content = c.pyapi.to_native_value(tpe.contenttpe, content_obj).value
    c.pyapi.decref(asarray_obj)
    c.pyapi.decref(index_obj)
    c.pyapi.decref(content_obj)
    c.pyapi.decref(indexarray_obj)
    if tpe.identitiestpe != numba.none:
        id_obj = c.pyapi.object_getattr_string(obj, "identities")
        proxyout.identities = c.pyapi.to_native_value(tpe.identitiestpe, id_obj).value
        c.pyapi.decref(id_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(IndexedArrayType)
def box(tpe, val, c):
    if tpe.indexname == "64":
        Index_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.Index64))
        if tpe.isoption:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedOptionArray64))
        else:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedArray64))
    elif tpe.indexname == "32":
        Index_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.Index32))
        if tpe.isoption:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedOptionArray32))
        else:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedArray32))
    elif tpe.indexname == "U32":
        Index_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexU32))
        if tpe.isoption:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedOptionArrayU32))
        else:
            IndexedArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.IndexedArrayU32))
    else:
        raise AssertionError("unrecognized index type: {0}".format(tpe.indexname))
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    indexarray_obj = c.pyapi.from_native_value(tpe.indextpe, proxyin.index, c.env_manager)
    content_obj = c.pyapi.from_native_value(tpe.contenttpe, proxyin.content, c.env_manager)
    index_obj = c.pyapi.call_function_objargs(Index_obj, (indexarray_obj,))
    c.pyapi.decref(Index_obj)
    c.pyapi.decref(indexarray_obj)
    args = [index_obj, content_obj]
    if tpe.identitiestpe != numba.none:
        args.append(c.pyapi.from_native_value(tpe.identitiestpe, proxyin.identities, c.env_manager))
    else:
        args.append(c.pyapi.make_none())
    args.append(util.parameters2dict_impl(c, tpe.parameters))
    out = c.pyapi.call_function_objargs(IndexedArray_obj, args)
    for x in args:
        c.pyapi.decref(x)
    c.pyapi.decref(IndexedArray_obj)
    return out

@numba.extending.lower_builtin(len, IndexedArrayType)
def lower_len(context, builder, sig, args):
    rettpe, (tpe,) = sig.return_type, sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    indexlen = numba.targets.arrayobj.array_len(context, builder, numba.intp(tpe.indextpe), (proxyin.index,))
    return indexlen

@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.Integer)
def lower_getitem_int(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)

    indexval = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, tpe.indextpe.dtype(tpe.indextpe, wheretpe), (proxyin.index, whereval))

    outval = tpe.contenttpe.lower_getitem_int(context, builder, tpe.contenttpe.getitem_int()(tpe.contenttpe, tpe.indextpe.dtype), (proxyin.content, indexval))

    if tpe.isoption:
        output = context.make_helper(builder, tpe.getitem_int())
        with builder.if_else(builder.icmp_signed("<", indexval, context.get_constant(tpe.indextpe.dtype, 0))) as (isnone, isvalid):
            with isnone:
                output.valid = numba.cgutils.false_bit
                output.data = numba.cgutils.get_null_value(output.data.type)
            with isvalid:
                output.valid = numba.cgutils.true_bit
                output.data = tpe.contenttpe.lower_getitem_int(context, builder, tpe.contenttpe.getitem_int()(tpe.contenttpe, tpe.indextpe.dtype), (proxyin.content, indexval))
        return output._getvalue()

    else:
        return tpe.contenttpe.lower_getitem_int(context, builder, tpe.contenttpe.getitem_int()(tpe.contenttpe, tpe.indextpe.dtype), (proxyin.content, indexval))

@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.slice2_type)
def lower_getitem_range(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args

    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)

    proxyout = numba.cgutils.create_struct_proxy(tpe)(context, builder)
    proxyout.index = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, tpe.indextpe(tpe.indextpe, wheretpe), (proxyin.index, whereval))
    proxyout.content = proxyin.content
    if tpe.identitiestpe != numba.none:
        proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, tpe.identitiestpe, wheretpe, proxyin.identities, whereval)

    out = proxyout._getvalue()
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, out)
    return out

@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.StringLiteral)
def lower_getitem_str(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args

    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    proxyout = numba.cgutils.create_struct_proxy(rettpe)(context, builder)
    proxyout.index = proxyin.index
    proxyout.content = tpe.contenttpe.lower_getitem_str(context, builder, rettpe.contenttpe(tpe.contenttpe, wheretpe), (proxyin.content, whereval))
    if tpe.identitiestpe != numba.none:
        proxyout.identities = proxyin.identities

    out = proxyout._getvalue()
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, out)
    return out

@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.BaseTuple)
def lower_getitem_tuple(context, builder, sig, args):
    return content.lower_getitem_tuple(context, builder, sig, args)

@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.Array)
@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.List)
@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.ArrayCompatible)
@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, numba.types.EllipsisType)
@numba.extending.lower_builtin(operator.getitem, IndexedArrayType, type(numba.typeof(numpy.newaxis)))
def lower_getitem_other(context, builder, sig, args):
    return content.lower_getitem_other(context, builder, sig, args)

def lower_getitem_next(context, builder, arraytpe, wheretpe, arrayval, whereval, advanced):
    if len(wheretpe.types) == 0:
        return arrayval

    headtpe = wheretpe.types[0]
    tailtpe = numba.types.Tuple(wheretpe.types[1:])
    headval = numba.cgutils.unpack_tuple(builder, whereval)[0]
    tailval = context.make_tuple(builder, tailtpe, numba.cgutils.unpack_tuple(builder, whereval)[1:])

    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)
    lenindex = util.arraylen(context, builder, arraytpe.indextpe, proxyin.index, totpe=numba.int64)
    lencontent = util.arraylen(context, builder, arraytpe.contenttpe, proxyin.content, totpe=numba.int64)

    if isinstance(headtpe, (numba.types.Integer, numba.types.SliceType, numba.types.Array)):
        if arraytpe.isoption:
            if arraytpe.indexname == "32":
                determine_numnull = cpu.kernels.awkward_indexedarray32_numnull
                kernel = cpu.kernels.awkward_indexedarray32_getitem_nextcarry_outindex_64
            elif arraytpe.indexname == "U32":
                determine_numnull = cpu.kernels.awkward_indexedarrayU32_numnull
                kernel = cpu.kernels.awkward_indexedarrayU32_getitem_nextcarry_outindex_64
            elif arraytpe.indexname == "64":
                determine_numnull = cpu.kernels.awkward_indexedarray64_numnull
                kernel = cpu.kernels.awkward_indexedarray64_getitem_nextcarry_outindex_64
            else:
                raise AssertionError("unrecognized index type: {0}".format(arraytpe.indexname))
            numnull = numba.cgutils.alloca_once(builder, context.get_value_type(numba.int64))
            util.call(context, builder, determine_numnull,
                (numnull,
                 util.arrayptr(context, builder, arraytpe.indextpe, proxyin.index),
                 context.get_constant(numba.int64, 0),
                 lenindex),
                "in {0}, indexing error".format(arraytpe.shortname))
            lencarry = builder.sub(lenindex, builder.load(numnull))
            nextcarry = util.newindex64(context, builder, numba.int64, lencarry)
            outindex = util.newindex(arraytpe.indexname, context, builder, numba.int64, lenindex)
            util.call(context, builder, kernel,
                (util.arrayptr(context, builder, util.index64tpe, nextcarry),
                 util.arrayptr(context, builder, util.indextpe(arraytpe.indexname), outindex),
                 util.arrayptr(context, builder, arraytpe.indextpe, proxyin.index),
                 context.get_constant(numba.int64, 0),
                 lenindex,
                 lencontent),
                "in {0}, indexing error".format(arraytpe.shortname))
            nextcontenttpe = arraytpe.contenttpe.carry()
            nextcontentval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)
            outret = nextcontenttpe.getitem_next(wheretpe, advanced is not None)
            outval = nextcontenttpe.lower_getitem_next(context, builder, nextcontenttpe, wheretpe, nextcontentval, whereval, advanced)
            rettpe = IndexedArrayType(arraytpe.indextpe, outret, arraytpe.isoption, arraytpe.identitiestpe, arraytpe.parameters)
            proxyout = numba.cgutils.create_struct_proxy(rettpe)(context, builder)
            proxyout.index = outindex
            proxyout.content = outval
            if arraytpe.identitiestpe != numba.none:
                proxyout.identities = proxyin.identities
            return proxyout._getvalue()

        else:
            if arraytpe.indexname == "32":
                kernel = cpu.kernels.awkward_indexedarray32_getitem_nextcarry_64
            elif arraytpe.indexname == "U32":
                kernel = cpu.kernels.awkward_indexedarrayU32_getitem_nextcarry_64
            elif arraytpe.indexname == "64":
                kernel = cpu.kernels.awkward_indexedarray64_getitem_nextcarry_64
            else:
                raise AssertionError("unrecognized index type: {0}".format(arraytpe.indexname))

            nextcarry = util.newindex64(context, builder, numba.int64, lenindex)
            util.call(context, builder, kernel,
                (util.arrayptr(context, builder, util.index64tpe, nextcarry),
                 util.arrayptr(context, builder, arraytpe.indextpe, proxyin.index),
                 context.get_constant(numba.int64, 0),
                 lenindex,
                 lencontent),
                "in {0}, indexing error".format(arraytpe.shortname))
            nextcontenttpe = arraytpe.contenttpe.carry()
            nextcontentval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)
            return nextcontenttpe.lower_getitem_next(context, builder, nextcontenttpe, wheretpe, nextcontentval, whereval, advanced)

    elif isinstance(headtpe, numba.types.StringLiteral):
        nexttpe = arraytpe.getitem_str(headtpe.literal_value)
        nextval = lower_getitem_str(context, builder, nexttpe(arraytpe, headtpe), (arrayval, headval))
        return lower_getitem_next(context, builder, nexttpe, tailtpe, nextval, tailval, advanced)

    elif isinstance(headtpe, numba.types.EllipsisType):
        raise NotImplementedError("ellipsis")

    elif isinstance(headtpe, type(numba.typeof(numpy.newaxis))):
        raise NotImplementedError("newaxis")

    else:
        raise AssertionError(headtpe)

def lower_carry(context, builder, arraytpe, carrytpe, arrayval, carryval):
    import awkward1._numba.identities

    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)

    proxyout = numba.cgutils.create_struct_proxy(arraytpe)(context, builder)
    proxyout.index = numba.targets.arrayobj.fancy_getitem_array(context, builder, arraytpe.indextpe(arraytpe.indextpe, carrytpe), (proxyin.index, carryval))
    proxyout.content = proxyin.content
    if arraytpe.identitiestpe != numba.none:
        proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, arraytpe.identitiestpe, carrytpe, proxyin.identities, carryval)

    return proxyout._getvalue()

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = IndexedArrayType

    def generic_resolve(self, tpe, attr):
        if attr == "index":
            return tpe.indextpe

        elif attr == "content":
            return tpe.contenttpe

        elif attr == "identities":
            if tpe.identitiestpe == numba.none:
                return numba.optional(identity.IdentitiesType(numba.int32[:, :]))
            else:
                return tpe.identitiestpe

@numba.extending.lower_getattr(IndexedArrayType, "index")
def lower_index(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if context.enable_nrt:
        context.nrt.incref(builder, tpe.indextpe, proxyin.index)
    return proxyin.index

@numba.extending.lower_getattr(IndexedArrayType, "content")
def lower_content(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if context.enable_nrt:
        context.nrt.incref(builder, tpe.contenttpe, proxyin.content)
    return proxyin.content

@numba.extending.lower_getattr(IndexedArrayType, "identities")
def lower_identities(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if tpe.identitiestpe == numba.none:
        return context.make_optional_none(builder, identity.IdentitiesType(numba.int32[:, :]))
    else:
        if context.enable_nrt:
            context.nrt.incref(builder, tpe.identitiestpe, proxyin.identities)
        return proxyin.identities



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]));
offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10, 10]))
listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content)
regulararray = awkward1.layout.RegularArray(listoffsetarray, 2)
starts = awkward1.layout.Index64(numpy.array([0, 1]))
stops = awkward1.layout.Index64(numpy.array([2, 3]))
listarray = awkward1.layout.ListArray64(starts, stops, regulararray)

def test_type():
    assert str(awkward1.typeof(regulararray)) == "2 * var * float64"

def test_iteration():
    assert awkward1.tolist(regulararray) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]

def test_tojson():
    assert awkward1.tojson(regulararray) == "[[[0.0,1.1,2.2],[]],[[3.3,4.4],[5.5]],[[6.6,7.7,8.8,9.9],[]]]"

def test_getitem_at():
    assert awkward1.tolist(regulararray[0]) == [[0.0, 1.1, 2.2], []]
    assert awkward1.tolist(regulararray[1]) == [[3.3, 4.4], [5.5]]
    assert awkward1.tolist(regulararray[2]) == [[6.6, 7.7, 8.8, 9.9], []]

def test_getitem_range():
    assert awkward1.tolist(regulararray[1:]) == [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    assert awkward1.tolist(regulararray[:-1]) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]

def test_getitem():
    assert awkward1.tolist(regulararray[(0,)]) == [[0.0, 1.1, 2.2], []]
    assert awkward1.tolist(regulararray[(1,)]) == [[3.3, 4.4], [5.5]]
    assert awkward1.tolist(regulararray[(2,)]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert awkward1.tolist(regulararray[(slice(1, None, None),)]) == [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    assert awkward1.tolist(regulararray[(slice(None, -1, None),)]) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]

def test_getitem_deeper():
    assert awkward1.tolist(listarray) == [[[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]], [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]]

    assert awkward1.tolist(listarray[0, 0, 0]) == [0.0, 1.1, 2.2]
    assert awkward1.tolist(listarray[0, 0, 1]) == []
    assert awkward1.tolist(listarray[0, 1, 0]) == [3.3, 4.4]
    assert awkward1.tolist(listarray[0, 1, 1]) == [5.5]
    assert awkward1.tolist(listarray[1, 0, 0]) == [3.3, 4.4]
    assert awkward1.tolist(listarray[1, 0, 1]) == [5.5]
    assert awkward1.tolist(listarray[1, 1, 0]) == [6.6, 7.7, 8.8, 9.9]
    assert awkward1.tolist(listarray[1, 1, 1]) == []

    assert awkward1.tolist(listarray[0, 0, 0:]) == [[0.0, 1.1, 2.2], []]
    assert awkward1.tolist(listarray[0, 0, 1:]) == [[]]
    assert awkward1.tolist(listarray[0, 1, 0:]) == [[3.3, 4.4], [5.5]]
    assert awkward1.tolist(listarray[0, 1, 1:]) == [[5.5]]
    assert awkward1.tolist(listarray[1, 0, 0:]) == [[3.3, 4.4], [5.5]]
    assert awkward1.tolist(listarray[1, 0, 1:]) == [[5.5]]
    assert awkward1.tolist(listarray[1, 1, 0:]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert awkward1.tolist(listarray[1, 1, 1:]) == [[]]

    assert awkward1.tolist(listarray[[1], 0, 0:]) == [[[3.3, 4.4], [5.5]]]
    assert awkward1.tolist(listarray[[1, 0], 0, 0:]) == [[[3.3, 4.4], [5.5]], [[0.0, 1.1, 2.2], []]]

    assert awkward1.tolist(listarray[:, :, [0, 1]]) == [[[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]], [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]]
    assert awkward1.tolist(listarray[:, :, [1, 0]]) == [[[[], [0.0, 1.1, 2.2]], [[5.5], [3.3, 4.4]]], [[[5.5], [3.3, 4.4]], [[], [6.6, 7.7, 8.8, 9.9]]]]
    assert awkward1.tolist(listarray[:, :, [1, 0, 1]]) == [[[[], [0.0, 1.1, 2.2], []], [[5.5], [3.3, 4.4], [5.5]]], [[[5.5], [3.3, 4.4], [5.5]], [[], [6.6, 7.7, 8.8, 9.9], []]]]
    assert awkward1.tolist(listarray[:, :2, [0, 1]]) == [[[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]], [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]]

    assert awkward1.tolist(listarray[:1, [0, 0, 1, 1], [0, 1, 0, 1]]) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5]]]
    assert awkward1.tolist(listarray[:1, [1, 1, 0, 0], [1, 0, 1, 0]]) == [[[5.5], [3.3, 4.4], [], [0.0, 1.1, 2.2]]]

content2 = awkward1.layout.NumpyArray(numpy.arange(2*3*5*7).reshape(-1, 7))
regulararrayA = awkward1.layout.RegularArray(content2, 5)
regulararrayB = awkward1.layout.RegularArray(regulararrayA, 3)
modelA = numpy.arange(2*3*5*7).reshape(2*3, 5, 7)
modelB = numpy.arange(2*3*5*7).reshape(2, 3, 5, 7)

def test_numpy():
    assert awkward1.tolist(regulararrayA) == awkward1.tolist(modelA)
    assert awkward1.tolist(regulararrayB) == awkward1.tolist(modelB)

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(1, None), slice(None, -1), 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(regulararrayA[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(regulararrayB[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, 1, slice(1, None), slice(None, -1)), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(regulararrayB[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(regulararrayB[cuts])

def test_setidentities():
    regulararray.setidentities()
    assert numpy.asarray(regulararray.identities).tolist() == [
        [0],
        [1],
        [2]]
    assert numpy.asarray(regulararray.content.identities).tolist() == [
        [0, 0],
        [0, 1],
        [1, 0],
        [1, 1],
        [2, 0],
        [2, 1]]
    assert numpy.asarray(regulararray.content.content.identities).tolist() == [
        [0, 0, 0],   # 0.0
        [0, 0, 1],   # 1.1
        [0, 0, 2],   # 2.2
      # [0, 1,  ],   # (empty list)
        [1, 0, 0],   # 3.3
        [1, 0, 1],   # 4.4
        [1, 1, 0],   # 5.5
        [2, 0, 0],   # 6.6
        [2, 0, 1],   # 7.7
        [2, 0, 2],   # 8.8
        [2, 0, 3]]   # 9.9
      # [2, 1,  ],   # (empty list)

    regulararrayB.setidentities()
    assert numpy.asarray(regulararrayB.identities).tolist() == [
        [0],
        [1]]
    assert numpy.asarray(regulararrayB.content.identities).tolist() == [
        [0, 0],
        [0, 1],
        [0, 2],
        [1, 0],
        [1, 1],
        [1, 2]]

numba = pytest.importorskip("numba")

def test_numba():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]));
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content)
    regulararray = awkward1.layout.RegularArray(listoffsetarray, 2)
    regulararray_m1 = awkward1.layout.RegularArray(listoffsetarray[:-1], 2)
    regulararray_m2 = awkward1.layout.RegularArray(listoffsetarray[:-2], 2)
    starts = awkward1.layout.Index64(numpy.array([0, 1]))
    stops = awkward1.layout.Index64(numpy.array([2, 3]))
    listarray = awkward1.layout.ListArray64(starts, stops, regulararray)

    @numba.njit
    def f1(q):
        return 3.14

    assert sys.getrefcount(regulararray) == 2
    f1(regulararray)
    assert sys.getrefcount(regulararray) == 2

    @numba.njit
    def f2(q):
        return q

    assert sys.getrefcount(regulararray) == 2
    assert awkward1.tolist(f2(regulararray)) == awkward1.tolist(regulararray)
    assert sys.getrefcount(regulararray) == 2

    @numba.njit
    def f3(q):
        return len(q)

    assert f3(regulararray) == 3
    assert f3(regulararray_m1) == 2
    assert f3(regulararray_m2) == 2
    assert len(regulararray) == 3
    assert len(regulararray_m1) == 2
    assert len(regulararray_m2) == 2

    @numba.njit
    def f4(q):
        return q[1]

    assert awkward1.tolist(f4(regulararray)) == [[3.3, 4.4], [5.5]]

    @numba.njit
    def f5(q, i):
        return q[i]

    assert awkward1.tolist(f5(regulararray, 1)) == [[3.3, 4.4], [5.5]]

    @numba.njit
    def f6(q, i):
        return q[i]

    assert awkward1.tolist(f6(regulararray, slice(1, None))) == awkward1.tolist(regulararray[slice(1, None)]) == [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    assert awkward1.tolist(f6(regulararray, slice(None, -1))) == awkward1.tolist(regulararray[slice(None, -1)]) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]

    @numba.njit
    def f7(q, i):
        return q[(i,)]

    assert awkward1.tolist(f7(regulararray, 1)) == [[3.3, 4.4], [5.5]]

    @numba.njit
    def f8(q, i):
        return q[:, i]

    assert awkward1.tolist(f8(listarray, 1)) == awkward1.tolist(listarray[:, 1]) == [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]

    @numba.njit
    def f9(q):
        return q[:, 1, 0]

    assert awkward1.tolist(listarray[:, 1, 0]) == awkward1.tolist(f9(listarray)) == [[3.3, 4.4], [6.6, 7.7, 8.8, 9.9]]

    @numba.njit
    def f10(q, i):
        return q[:, 1, i]

    assert awkward1.tolist(f10(listarray, slice(None, 1))) == awkward1.tolist(listarray[:, 1, slice(None, 1)]) == [[[3.3, 4.4]], [[6.6, 7.7, 8.8, 9.9]]]
    assert awkward1.tolist(f10(listarray, slice(None, None))) == [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    assert awkward1.tolist(f10(listarray, slice(None, None, -1))) == [[[5.5], [3.3, 4.4]], [[], [6.6, 7.7, 8.8, 9.9]]]

    @numba.njit
    def f11(q, i):
        return q[[0, 1], 1, i]

    assert awkward1.tolist(f11(listarray, slice(None, 1))) == awkward1.tolist(listarray[[0, 1], 1, slice(None, 1)]) == [[[3.3, 4.4]], [[6.6, 7.7, 8.8, 9.9]]]

    @numba.njit
    def f12(q):
        return q[:, 1:, [1, 0, 1]]

    assert awkward1.tolist(f12(listarray)) == awkward1.tolist(listarray[:, 1:, [1, 0, 1]]) == [[[[5.5], [3.3, 4.4], [5.5]]], [[[], [6.6, 7.7, 8.8, 9.9], []]]]

    @numba.njit
    def f13(q):
        return q[:, [1], [1, 0, 1]]

    assert awkward1.tolist(f13(listarray)) == awkward1.tolist(listarray[:, [1], [1, 0, 1]]) == [[[5.5], [3.3, 4.4], [5.5]], [[], [6.6, 7.7, 8.8, 9.9], []]]

    @numba.njit
    def f14(q):
        return q.content

    assert awkward1.tolist(f14(regulararray)) == awkward1.tolist(regulararray.content)

    @numba.njit
    def f15(q):
        return q.size

    assert awkward1.tolist(f15(regulararray)) == awkward1.tolist(regulararray.size) == 2



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba
import numba.typing.arraydecl
import numba.typing.ctypes_utils

import awkward1.layout
from ..._numba import cpu, util, identities, content

@numba.extending.typeof_impl.register(awkward1.layout.NumpyArray)
def typeof(val, c):
    import awkward1._numba.types
    type = val.type
    while isinstance(type, (awkward1.layout.ArrayType, awkward1.layout.RegularType)):
        type = type.type
    return NumpyArrayType(numba.typeof(numpy.asarray(val)), numba.typeof(val.identities), util.dict2parameters(val.parameters))

class NumpyArrayType(content.ContentType):
    def __init__(self, arraytpe, identitiestpe, parameters):
        assert isinstance(parameters, tuple)
        super(NumpyArrayType, self).__init__(name="ak::NumpyArrayType({0}, identities={1}, parameters={2})".format(arraytpe.name, identitiestpe.name, util.parameters2str(parameters)))
        self.arraytpe = arraytpe
        self.identitiestpe = identitiestpe
        self.parameters = parameters

    @property
    def ndim(self):
        return self.arraytpe.ndim

    def getitem_int(self):
        return self.getitem_tuple(numba.int64)

    def getitem_range(self):
        return self.getitem_tuple(numba.types.slice2_type)

    def getitem_str(self):
        raise IndexError("cannot slice NumpyArray with str (Record field name)")

    def getitem_tuple(self, wheretpe):
        outtpe = numba.typing.arraydecl.get_array_index_type(self.arraytpe, wheretpe).result
        if isinstance(outtpe, numba.types.Array):
            return NumpyArrayType(outtpe, self.identitiestpe, self.parameters)
        else:
            return outtpe

    def getitem_next(self, wheretpe, isadvanced):
        if len(wheretpe.types) > self.arraytpe.ndim:
            raise IndexError("too many dimensions in slice")
        if any(isinstance(x, numba.types.StringLiteral) for x in wheretpe):
            raise IndexError("cannot slice NumpyArray with str (Record field name)")

        if isadvanced:
            numreduce = sum(1 if isinstance(x, (numba.types.Integer, numba.types.Array)) else 0 for x in wheretpe.types)
        else:
            numreduce = sum(1 if isinstance(x, numba.types.Integer) else 0 for x in wheretpe.types)
        if numreduce < self.arraytpe.ndim:
            return NumpyArrayType(numba.types.Array(self.arraytpe.dtype, self.arraytpe.ndim - numreduce, self.arraytpe.layout), self.identitiestpe, self.parameters)
        elif numreduce == self.arraytpe.ndim:
            return self.arraytpe.dtype
        else:
            assert False

    def carry(self):
        return self

    @property
    def lower_len(self):
        return lower_len

    @property
    def lower_getitem_nothing(self):
        return None

    @property
    def lower_getitem_int(self):
        return lower_getitem

    @property
    def lower_getitem_range(self):
        return lower_getitem

    @property
    def lower_getitem_next(self):
        return lower_getitem_next

    @property
    def lower_carry(self):
        return lower_carry

@numba.extending.register_model(NumpyArrayType)
class NumpyArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("array", fe_type.arraytpe)]
        if fe_type.identitiestpe != numba.none:
            members.append(("identities", fe_type.identitiestpe))
        super(NumpyArrayModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(NumpyArrayType)
def unbox(tpe, obj, c):
    asarray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(numpy.asarray))
    array_obj = c.pyapi.call_function_objargs(asarray_obj, (obj,))
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    proxyout.array = c.pyapi.to_native_value(tpe.arraytpe, array_obj).value
    c.pyapi.decref(asarray_obj)
    c.pyapi.decref(array_obj)
    if tpe.identitiestpe != numba.none:
        id_obj = c.pyapi.object_getattr_string(obj, "identities")
        proxyout.identities = c.pyapi.to_native_value(tpe.identitiestpe, id_obj).value
        c.pyapi.decref(id_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(NumpyArrayType)
def box(tpe, val, c):
    NumpyArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.NumpyArray))
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    array_obj = c.pyapi.from_native_value(tpe.arraytpe, proxyin.array, c.env_manager)
    args = [array_obj]
    if tpe.identitiestpe != numba.none:
        args.append(c.pyapi.from_native_value(tpe.identitiestpe, proxyin.identities, c.env_manager))
    else:
        args.append(c.pyapi.make_none())
    args.append(util.parameters2dict_impl(c, tpe.parameters))
    out = c.pyapi.call_function_objargs(NumpyArray_obj, args)
    for x in args:
        c.pyapi.decref(x)
    c.pyapi.decref(NumpyArray_obj)
    return out

@numba.extending.lower_builtin(len, NumpyArrayType)
def lower_len(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    return numba.targets.arrayobj.array_len(context, builder, numba.intp(tpe.arraytpe), (proxyin.array,))

@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.Integer)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.SliceType)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.Array)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.List)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.ArrayCompatible)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.EllipsisType)
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, type(numba.typeof(numpy.newaxis)))
@numba.extending.lower_builtin(operator.getitem, NumpyArrayType, numba.types.BaseTuple)
def lower_getitem(context, builder, sig, args):
    import awkward1._numba.identities

    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)

    if not isinstance(wheretpe, (numba.types.Integer, numba.types.SliceType)):
        if not isinstance(wheretpe, numba.types.BaseTuple):
            wheretpe = numba.types.Tuple((wheretpe,))
            whereval = context.make_tuple(builder, wheretpe, (whereval,))
        wheretpe, whereval = util.preprocess_slicetuple(context, builder, wheretpe, whereval)

    if isinstance(rettpe, NumpyArrayType):
        signature = rettpe.arraytpe(tpe.arraytpe, wheretpe)
    else:
        signature = rettpe(tpe.arraytpe, wheretpe)

    if isinstance(wheretpe, numba.types.BaseTuple):
        out = numba.targets.arrayobj.getitem_array_tuple(context, builder, signature, (proxyin.array, whereval))
    else:
        out = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, signature, (proxyin.array, whereval))

    if isinstance(rettpe, NumpyArrayType):
        proxyout = numba.cgutils.create_struct_proxy(rettpe)(context, builder)
        proxyout.array = out
        if rettpe.identitiestpe != numba.none:
            proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, rettpe.identitiestpe, wheretpe, proxyin.identities, whereval)
        return proxyout._getvalue()
    else:
        return out

def lower_getitem_next(context, builder, arraytpe, wheretpe, arrayval, whereval, advanced):
    if len(wheretpe.types) == 0:
        return arrayval
    headtpe = wheretpe.types[0]
    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)

    if isinstance(headtpe, numba.types.Array) and advanced is not None:
        tailtpe = numba.types.Tuple(wheretpe.types[1:])
        headval = numba.cgutils.unpack_tuple(builder, whereval)[0]
        tailval = context.make_tuple(builder, tailtpe, numba.cgutils.unpack_tuple(builder, whereval)[1:])
        if headtpe.ndim != 1:
            raise NotImplementedError("array.ndim != 1")
        if arraytpe.arraytpe.ndim < 2:
            raise TypeError("too many dimensions in slice")

        shapeval = numba.targets.arrayobj.make_array(arraytpe.arraytpe)(context, builder, proxyin.array).shape
        shapeunpacked = numba.cgutils.unpack_tuple(builder, shapeval)
        lenself, skip = shapeunpacked[:2]
        lennext = builder.mul(lenself, skip)

        carry = util.newindex64(context, builder, numba.intp, lenself)
        util.call(context, builder, cpu.kernels.awkward_carry_arange_64,
            (util.arrayptr(context, builder, util.index64tpe, carry),
             util.cast(context, builder, numba.intp, numba.int64, lenself)),
            "in {0}, indexing error".format(arraytpe.shortname))

        flathead = numba.targets.arrayobj.array_flatten(context, builder, util.index64tpe(headtpe), (headval,))
        lenflathead = util.arraylen(context, builder, util.index64tpe, flathead)

        util.call(context, builder, cpu.kernels.awkward_regularize_arrayslice_64,
            (util.arrayptr(context, builder, util.index64tpe, flathead),
             util.cast(context, builder, numba.intp, numba.int64, lenflathead),
             util.cast(context, builder, numba.intp, numba.int64, skip)),
            "in {0}, indexing error".format(arraytpe.shortname))

        nextcarry = util.newindex64(context, builder, numba.intp, lenself)
        util.call(context, builder, cpu.kernels.awkward_numpyarray_getitem_next_array_advanced_64,
            (util.arrayptr(context, builder, util.index64tpe, nextcarry),
             util.arrayptr(context, builder, util.index64tpe, carry),
             util.arrayptr(context, builder, util.index64tpe, advanced),
             util.arrayptr(context, builder, util.index64tpe, flathead),
             util.cast(context, builder, numba.intp, numba.int64, lenself),
             util.cast(context, builder, numba.intp, numba.int64, skip)),
            "in {0}, indexing error".format(arraytpe.shortname))

        nextshapetpe = numba.types.UniTuple(numba.intp, arraytpe.arraytpe.ndim - 1)
        nextshapeval = context.make_tuple(builder, nextshapetpe, [lennext] + shapeunpacked[2:])

        nextarraytpe = numba.types.Array(arraytpe.arraytpe.dtype, arraytpe.arraytpe.ndim - 1, arraytpe.arraytpe.layout)
        nextarrayval = numba.targets.arrayobj.array_reshape(context, builder, nextarraytpe(arraytpe.arraytpe, nextshapetpe), (proxyin.array, nextshapeval))

        nexttpe = NumpyArrayType(nextarraytpe, arraytpe.identitiestpe, arraytpe.parameters)
        proxynext = numba.cgutils.create_struct_proxy(nexttpe)(context, builder)
        proxynext.array = nextarrayval
        if arraytpe.identitiestpe != numba.none:
            proxynext.identities = proxyin.identities

        outval = lower_carry(context, builder, nexttpe, util.index64tpe, proxynext._getvalue(), nextcarry)
        return lower_getitem_next(context, builder, nexttpe, tailtpe, outval, tailval, advanced)

    else:
        proxyslice = numba.cgutils.create_struct_proxy(numba.types.slice2_type)(context, builder)
        proxyslice.start = context.get_constant(numba.intp, 0)
        proxyslice.stop = util.arraylen(context, builder, arraytpe.arraytpe, proxyin.array, totpe=numba.intp)
        proxyslice.step = context.get_constant(numba.intp, 1)
        wheretpe = numba.types.Tuple((numba.types.slice2_type,) + wheretpe.types)
        whereval = context.make_tuple(builder, wheretpe, [proxyslice._getvalue()] + numba.cgutils.unpack_tuple(builder, whereval))

        outtpe = numba.typing.arraydecl.get_array_index_type(arraytpe.arraytpe, wheretpe).result
        outval = numba.targets.arrayobj.getitem_array_tuple(context, builder, outtpe(arraytpe.arraytpe, wheretpe), (proxyin.array, whereval))

        if isinstance(outtpe, numba.types.Array):
            proxyout = numba.cgutils.create_struct_proxy(NumpyArrayType(outtpe, arraytpe.identitiestpe, arraytpe.parameters))(context, builder)
            proxyout.array = outval
            if arraytpe.identitiestpe != numba.none:
                proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, arraytpe.identitiestpe, wheretpe, proxyin.identities, whereval)
            return proxyout._getvalue()
        else:
            return out

def lower_carry(context, builder, arraytpe, carrytpe, arrayval, carryval):
    import awkward1._numba.identities

    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)

    proxyout = numba.cgutils.create_struct_proxy(arraytpe)(context, builder)
    proxyout.array = numba.targets.arrayobj.fancy_getitem_array(context, builder, arraytpe.arraytpe(arraytpe.arraytpe, carrytpe), (proxyin.array, carryval))
    if arraytpe.identitiestpe != numba.none:
        proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, arraytpe.identitiestpe, carrytpe, proxyin.identities, carryval)
    return proxyout._getvalue()

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = NumpyArrayType

    def generic_resolve(self, tpe, attr):
        if attr == "identities":
            if tpe.identitiestpe == numba.none:
                return numba.optional(identities.IdentitiesType(numba.int32[:, :]))
            else:
                return tpe.identitiestpe

@numba.extending.lower_getattr(NumpyArrayType, "identities")
def lower_identities(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if tpe.identitiestpe == numba.none:
        return context.make_optional_none(builder, identities.IdentitiesType(numba.int32[:, :]))
    else:
        if context.enable_nrt:
            context.nrt.incref(builder, tpe.identitiestpe, proxyin.identities)
        return proxyin.identities



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import distutils.version

import numpy
if distutils.version.LooseVersion(numpy.__version__) < distutils.version.LooseVersion("1.13.1"):
    raise ImportError("Numpy 1.13.1 or later required")

classes = {}
functions = {}

import awkward1.layout
from awkward1.layout import Type
from awkward1.layout import UnknownType
from awkward1.layout import PrimitiveType
from awkward1.layout import ListType
from awkward1.layout import RegularType
from awkward1.layout import RecordType
from awkward1.layout import OptionType
from awkward1.layout import UnionType
from awkward1.layout import ArrayType

import awkward1._numba

import awkward1.highlevel
from awkward1.highlevel import Array
from awkward1.highlevel import Record
from awkward1.highlevel import FillableArray

from awkward1.operations.convert import *
from awkward1.operations.describe import *
from awkward1.operations.structure import *

from awkward1.behavior.string import *

__version__ = awkward1.layout.__version__



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import os
import platform
import re
import subprocess
import sys
import shutil
import glob

import distutils.version
import setuptools
import setuptools.command.build_ext
from setuptools import setup

class CMakeExtension(setuptools.Extension):
    def __init__(self, name, sourcedir=""):
        setuptools.Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(setuptools.command.build_ext.build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " + ", ".join(x.name for x in self.extensions))

        if platform.system() == "Windows":
            cmake_version = distutils.version.LooseVersion(re.search(r"version\s*([\d.]+)", out.decode()).group(1))
            if cmake_version < "3.4":
                raise RuntimeError("CMake >= 3.4 is required on Windows")

        for x in self.extensions:
            self.build_extension(x)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir, "-DPYTHON_EXECUTABLE=" + sys.executable]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{0}={1}".format(cfg.upper(), extdir), "-DCMAKE_WINDOWS_EXPORT_ALL_SYMBOLS=TRUE"]
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]

        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp)
        subprocess.check_call(["cmake", "--build", "."] + build_args, cwd=self.build_temp)
        subprocess.check_call(["ctest", "--output-on-failure"], cwd=self.build_temp)

        for lib in (glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "libawkward-cpu-kernels-static.*")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "libawkward-static.*")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "*.so")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "*.dylib")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "*.dll")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "*.exp")) +
                    glob.glob(os.path.join(os.path.join(extdir, "awkward1"), "*.pyd"))):
          if os.path.exists(lib):
              os.remove(lib)

        for lib in os.listdir(self.build_temp):
            if lib.startswith("libawkward-cpu-kernels-static.") or lib.startswith("libawkward-static."):
                shutil.copy(os.path.join(self.build_temp, lib), "awkward1")
                shutil.move(os.path.join(self.build_temp, lib), os.path.join(extdir, "awkward1"))

        for lib in os.listdir(extdir):
            if lib.endswith(".so") or lib.endswith(".dylib") or lib.endswith(".dll") or lib.endswith(".pyd"):
                shutil.copy(os.path.join(extdir, lib), "awkward1")
                shutil.move(os.path.join(extdir, lib), os.path.join(extdir, "awkward1"))

        if platform.system() == "Windows":
            for lib in os.listdir(os.path.join(self.build_temp, cfg)):
                if lib.startswith("awkward-cpu-kernels-static.") or lib.startswith("awkward-static.") or lib.endswith(".dll") or lib.endswith(".exp") or lib.endswith(".pyd"):
                    shutil.copy(os.path.join(os.path.join(self.build_temp, cfg), lib), "awkward1")
                    shutil.move(os.path.join(os.path.join(self.build_temp, cfg), lib), os.path.join(extdir, "awkward1"))

setup(name = "awkward1",
      packages = setuptools.find_packages(exclude=["tests"]),
      scripts = [],
      data_files = [(os.path.join("awkward1", "signatures"), [x for x in glob.glob(os.path.join(os.path.join("awkward1", "signatures"), "*.xml")) if x != "index.xml" and not x.startswith("dir_") and not x.startswith("namespace")])],
      version = open("VERSION_INFO").read().strip(),
      author = "Jim Pivarski",
      author_email = "pivarski@princeton.edu",
      maintainer = "Jim Pivarski",
      maintainer_email = "pivarski@princeton.edu",
      description = "Development of awkward 1.0, to replace scikit-hep/awkward-array in 2020.",
      long_description = "",
      long_description_content_type = "text/markdown",
      url = "https://github.com/jpivarski/awkward1",
      download_url = "https://github.com/jpivarski/awkward1/releases",
      license = "BSD 3-clause",
      ext_modules = [CMakeExtension("awkward")],
      cmdclass = {"build_ext": CMakeBuild},
      test_suite = "tests",
      python_requires = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
      install_requires = open("requirements.txt").read().strip().split(),
      tests_require = open("requirements-test.txt").read().strip().split(),
      zip_safe = False,
      classifiers = [
#         "Development Status :: 1 - Planning",
#         "Development Status :: 2 - Pre-Alpha",
          "Development Status :: 3 - Alpha",
#         "Development Status :: 4 - Beta",
#         "Development Status :: 5 - Production/Stable",
#         "Development Status :: 6 - Mature",
#         "Development Status :: 7 - Inactive",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      )



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba

from .._numba import util

class ContentType(numba.types.Type):
    @property
    def shortname(self):
        return self.name[:self.name.index("Type")]

@numba.typing.templates.infer_global(len)
class type_len(numba.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            arraytpe, = args
            if isinstance(arraytpe, ContentType):
                return numba.typing.templates.signature(numba.types.intp, arraytpe)

@numba.typing.templates.infer_global(operator.getitem)
class type_getitem(numba.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 2 and len(kwargs) == 0:
            arraytpe, wheretpe = args

            if isinstance(arraytpe, ContentType):
                original_wheretpe = wheretpe
                if isinstance(wheretpe, numba.types.Integer):
                    return numba.typing.templates.signature(arraytpe.getitem_int(), arraytpe, original_wheretpe)
                if isinstance(wheretpe, numba.types.SliceType) and not wheretpe.has_step:
                    return numba.typing.templates.signature(arraytpe.getitem_range(), arraytpe, original_wheretpe)
                if isinstance(wheretpe, numba.types.StringLiteral):
                    return numba.typing.templates.signature(arraytpe.getitem_str(wheretpe.literal_value), arraytpe, original_wheretpe)

                if not isinstance(wheretpe, numba.types.BaseTuple):
                    wheretpe = numba.types.Tuple((wheretpe,))

                wheretpe = util.typing_regularize_slice(wheretpe)
                self.check_slice_types(wheretpe)

                return numba.typing.templates.signature(arraytpe.getitem_tuple(wheretpe), arraytpe, original_wheretpe)

    @staticmethod
    def check_slice_types(wheretpe):
        if any(not isinstance(t, (numba.types.Integer, numba.types.SliceType, numba.types.EllipsisType, type(numba.typeof(numpy.newaxis)), numba.types.StringLiteral)) and not (isinstance(t, numba.types.Array) and isinstance(t.dtype, (numba.types.Boolean, numba.types.Integer))) for t in wheretpe.types):
            raise TypeError("only integers, slices (`:`), ellipsis (`...`), numpy.newaxis (`None`), integer or boolean arrays (possibly jagged), and constant strings (known at compile-time) are valid indices")


def lower_getitem_nothing(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    proxyslice = numba.cgutils.create_struct_proxy(numba.types.slice2_type)(context, builder)
    proxyslice.start = context.get_constant(numba.intp, 0)
    proxyslice.stop = context.get_constant(numba.intp, 0)
    proxyslice.step = context.get_constant(numba.intp, 1)
    outtpe = tpe.contenttpe.getitem_range()
    return tpe.contenttpe.lower_getitem_range(context, builder, outtpe(tpe.contenttpe, numba.types.slice2_type), (proxyin.content, proxyslice._getvalue()))

def lower_getitem_tuple(context, builder, sig, args):
    import awkward1._numba.array.listarray
    import awkward1._numba.array.regulararray

    rettpe, (arraytpe, wheretpe) = sig.return_type, sig.args
    arrayval, whereval = args

    wheretpe, whereval = util.preprocess_slicetuple(context, builder, wheretpe, whereval)

    nexttpe = awkward1._numba.array.regulararray.RegularArrayType(arraytpe, numba.none, ())
    outtpe = nexttpe.getitem_next(wheretpe, False)
    if outtpe.lower_getitem_nothing is None:
        nexttpe = awkward1._numba.array.listarray.ListArrayType(util.index64tpe, util.index64tpe, arraytpe, numba.none, ())
        outtpe = nexttpe.getitem_next(wheretpe, False)

        length = util.arraylen(context, builder, arraytpe, arrayval, totpe=numba.int64)
        proxynext = numba.cgutils.create_struct_proxy(nexttpe)(context, builder)
        proxynext.starts = util.newindex64(context, builder, numba.int64, context.get_constant(numba.int64, 1))
        proxynext.stops = util.newindex64(context, builder, numba.int64, context.get_constant(numba.int64, 1))
        numba.targets.arrayobj.store_item(context, builder, util.index64tpe, context.get_constant(numba.int64, 0), util.arrayptr(context, builder, util.index64tpe, proxynext.starts))
        numba.targets.arrayobj.store_item(context, builder, util.index64tpe, length, util.arrayptr(context, builder, util.index64tpe, proxynext.stops))
        proxynext.content = arrayval
        nextval = proxynext._getvalue()

        outval = nexttpe.lower_getitem_next(context, builder, nexttpe, wheretpe, nextval, whereval, None)

        return outtpe.lower_getitem_int(context, builder, rettpe(outtpe, numba.int64), (outval, context.get_constant(numba.int64, 0)))

    else:
        proxynext = numba.cgutils.create_struct_proxy(nexttpe)(context, builder)
        proxynext.content = arrayval
        proxynext.size = util.arraylen(context, builder, arraytpe, arrayval, totpe=numba.int64)
        nextval = proxynext._getvalue()

        outval = nexttpe.lower_getitem_next(context, builder, nexttpe, wheretpe, nextval, whereval, None)

        lenout = util.arraylen(context, builder, outtpe, outval)
        outputptr = numba.cgutils.alloca_once(builder, context.get_value_type(outtpe.getitem_int()))
        with builder.if_else(builder.icmp_signed("==", lenout, context.get_constant(numba.intp, 0)), likely=False) as (nothing, something):
            with nothing:
                builder.store(outtpe.lower_getitem_nothing(context, builder, outtpe, outval), outputptr)
            with something:
                builder.store(outtpe.lower_getitem_int(context, builder, rettpe(outtpe, numba.int64), (outval, context.get_constant(numba.int64, 0))), outputptr)
        return builder.load(outputptr)

def lower_getitem_other(context, builder, sig, args):
    rettpe, (arraytpe, wheretpe) = sig.return_type, sig.args
    arrayval, whereval = args
    wrappedtpe = numba.types.Tuple((wheretpe,))
    wrappedval = context.make_tuple(builder, wrappedtpe, (whereval,))
    return lower_getitem_tuple(context, builder, rettpe(arraytpe, wrappedtpe), (arrayval, wrappedval))



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import numpy

import awkward1

def test_refcount():
    o = numpy.arange(10, dtype="i4")
    c = numpy.arange(12).reshape(3, 4)

    for order in itertools.permutations(["del i, n", "del l", "del l2"]):
        i = awkward1.layout.Index32(o)
        n = awkward1.layout.NumpyArray(c)
        l = awkward1.layout.ListOffsetArray32(i, n)
        l2 = awkward1.layout.ListOffsetArray32(i, l)

        for statement in order:
            assert sys.getrefcount(o), sys.getrefcount(c) == (3, 3)
            exec(statement)
            assert sys.getrefcount(o), sys.getrefcount(c) == (2, 2)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import numbers

import numpy

import awkward1.layout

def parameters(array):
    if isinstance(array, (awkward1.highlevel.Array, awkward1.highlevel.Record)):
        return array.layout.parameters

    elif isinstance(array, (awkward1.layout.Content, awkward1.layout.Record)):
        return array.parameters

    elif isinstance(array, awkward1.highlevel.FillableArray):
        return array.snapshot().layout.parameters

    elif isinstance(array, awkward1.layout.FillableArray):
        return array.snapshot().parameters

    else:
        return {}

def typeof(array):
    if array is None:
        return awkward1.layout.UnknownType()

    elif isinstance(array, (bool, numpy.bool, numpy.bool_)):
        return awkward1.layout.PrimitiveType("bool")

    elif isinstance(array, numbers.Integral):
        return awkward1.layout.PrimitiveType("int64")

    elif isinstance(array, numbers.Real):
        return awkward1.layout.PrimitiveType("float64")

    elif isinstance(array, (numpy.int8, numpy.int16, numpy.int32, numpy.int64, numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64, numpy.float32, numpy.float64)):
        return awkward1.layout.PrimitiveType(typeof.dtype2primitive[array.dtype.type])

    elif isinstance(array, numpy.generic):
        raise ValueError("cannot describe {0} as a PrimitiveType".format(type(array)))

    elif isinstance(array, (awkward1.highlevel.Array, awkward1.highlevel.Record, awkward1.highlevel.FillableArray)):
        return array.type

    elif isinstance(array, awkward1.layout.Record):
        return array.type

    elif isinstance(array, numpy.ndarray):
        if len(array.shape) == 0:
            return typeof(array.reshape((1,))[0])
        else:
            out = awkward1.layout.PrimitiveType(typeof.dtype2primitive[array.dtype.type])
            for x in array.shape[-1:0:-1]:
                out = awkward1.layout.RegularType(out, x)
            return awkward1.layout.ArrayType(out, array.shape[0])

    elif isinstance(array, awkward1.layout.FillableArray):
        return array.type

    elif isinstance(array, awkward1.layout.Content):
        return array.type

    else:
        raise TypeError("unrecognized array type: {0}".format(repr(array)))

typeof.dtype2primitive = {
    numpy.int8:    "int8",
    numpy.int16:   "int16",
    numpy.int32:   "int32",
    numpy.int64:   "int64",
    numpy.uint8:   "uint8",
    numpy.uint16:  "uint16",
    numpy.uint32:  "uint32",
    numpy.uint64:  "uint64",
    numpy.float32: "float32",
    numpy.float64: "float64",
}



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

def test_empty_array_slice():
    # inspired by PR021::test_getitem
    a = awkward1.fromjson("[[], [[], []], [[], [], []]]")
    assert awkward1.tolist(a[2, 1, numpy.array([], dtype=int)]) == []
    assert awkward1.tolist(a[2, numpy.array([1], dtype=int), numpy.array([], dtype=int)]) == []

    # inspired by PR015::test_deep_numpy
    content = awkward1.layout.NumpyArray(numpy.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]))
    listarray = awkward1.layout.ListArray64(awkward1.layout.Index64(numpy.array([0, 3, 3])), awkward1.layout.Index64(numpy.array([3, 3, 5])), content)
    assert awkward1.tolist(listarray[[2, 0, 0, -1], [1, -1, 0, 0], [0, 1, 0, 1]]) == [8.8, 5.5, 0.0, 7.7]
    assert awkward1.tolist(listarray[2, 1, numpy.array([], dtype=int)]) == []
    assert awkward1.tolist(listarray[2, 1, []]) == []
    assert awkward1.tolist(listarray[2, [1], []]) == []
    assert awkward1.tolist(listarray[2, [], []]) == []

def test_nonflat_slice():
    array = numpy.arange(2*3*5).reshape(2, 3, 5)
    numpyarray = awkward1.layout.NumpyArray(array)

    content = awkward1.layout.NumpyArray(array.reshape(-1))
    inneroffsets = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30]))
    outeroffsets = awkward1.layout.Index64(numpy.array([0, 3, 6]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(outeroffsets, awkward1.layout.ListOffsetArray64(inneroffsets, content))
    listoffsetarray.setidentities()

    assert awkward1.tolist(array[[1, 0, 1, 1, 1, 0], [2, 0, 1, 1, 2, 0], [2, 4, 2, 4, 0, 1]]) == [27, 4, 22, 24, 25, 1]
    assert awkward1.tolist(array[[[1, 0], [1, 1], [1, 0]], [[2, 0], [1, 1], [2, 0]], [[2, 4], [2, 4], [0, 1]]]) == [[27, 4], [22, 24], [25, 1]]

    one = listoffsetarray[[1, 0, 1, 1, 1, 0], [2, 0, 1, 1, 2, 0], [2, 4, 2, 4, 0, 1]]
    assert awkward1.tolist(one) == [27, 4, 22, 24, 25, 1]
    assert numpy.asarray(one.identities).tolist() == [
        [1, 2, 2],
        [0, 0, 4],
        [1, 1, 2],
        [1, 1, 4],
        [1, 2, 0],
        [0, 0, 1]]

    two = listoffsetarray[[[1, 0], [1, 1], [1, 0]], [[2, 0], [1, 1], [2, 0]], [[2, 4], [2, 4], [0, 1]]]
    assert awkward1.tolist(two) == [[27, 4], [22, 24], [25, 1]]
    assert numpy.asarray(two.content.identities).tolist() == [
        [1, 2, 2],
        [0, 0, 4],
        [1, 1, 2],
        [1, 1, 4],
        [1, 2, 0],
        [0, 0, 1]]
    assert two.identities is None

def test_newaxis():
    array = numpy.arange(2*3*5).reshape(2, 3, 5)
    numpyarray = awkward1.layout.NumpyArray(array)

    content = awkward1.layout.NumpyArray(array.reshape(-1))
    inneroffsets = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30]))
    outeroffsets = awkward1.layout.Index64(numpy.array([0, 3, 6]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(outeroffsets, awkward1.layout.ListOffsetArray64(inneroffsets, content))

    assert awkward1.tolist(array[:, numpy.newaxis]) == [[[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]], [[[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]]

    assert awkward1.tolist(listoffsetarray[:, numpy.newaxis]) == [[[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]], [[[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]]

numba = pytest.importorskip("numba")

def test_empty_array_slice_numba():
    # inspired by PR015::test_deep_numpy
    content = awkward1.layout.NumpyArray(numpy.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]))
    listarray = awkward1.layout.ListArray64(awkward1.layout.Index64(numpy.array([0, 3, 3])), awkward1.layout.Index64(numpy.array([3, 3, 5])), content)

    @numba.njit
    def f1(q, i, j):
        return q[2, i, j]

    assert awkward1.tolist(f1(listarray, numpy.array([1], dtype=int), numpy.array([], dtype=int))) == []
    assert awkward1.tolist(f1(listarray, numpy.array([], dtype=int), numpy.array([], dtype=int))) == []



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba
import numba.typing.arraydecl
import numba.typing.ctypes_utils

import awkward1.layout
from .._numba import util

@numba.extending.typeof_impl.register(awkward1.layout.Identities32)
@numba.extending.typeof_impl.register(awkward1.layout.Identities64)
def typeof(val, c):
    return IdentitiesType(numba.typeof(val.array))

class IdentitiesType(numba.types.Type):
    fieldloctpe = numba.types.List(numba.types.Tuple((numba.int64, numba.types.string)))

    def bitwidth(self):
        return self.arraytpe.dtype.bitwidth

    def __init__(self, arraytpe):
        super(IdentitiesType, self).__init__(name="ak::Identities{0}Type".format(arraytpe.dtype.bitwidth))
        self.arraytpe = arraytpe

@numba.extending.register_model(IdentitiesType)
class IdentitiesModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("ref", util.RefType),
                   ("fieldloc", fe_type.fieldloctpe),
                   ("array", fe_type.arraytpe)]
        super(IdentitiesModel, self).__init__(dmm, fe_type, members)

numba.extending.make_attribute_wrapper(IdentitiesType, "ref", "ref")
numba.extending.make_attribute_wrapper(IdentitiesType, "fieldloc", "fieldloc")
numba.extending.make_attribute_wrapper(IdentitiesType, "array", "array")

@numba.extending.unbox(IdentitiesType)
def unbox(tpe, obj, c):
    ref_obj = c.pyapi.object_getattr_string(obj, "ref")
    fieldloc_obj = c.pyapi.object_getattr_string(obj, "fieldloc")
    array_obj = c.pyapi.object_getattr_string(obj, "array")
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    proxyout.ref = c.pyapi.to_native_value(util.RefType, ref_obj).value
    proxyout.fieldloc = c.pyapi.to_native_value(tpe.fieldloctpe, fieldloc_obj).value
    proxyout.array = c.pyapi.to_native_value(tpe.arraytpe, array_obj).value
    if c.context.enable_nrt:
        c.context.nrt.incref(c.builder, tpe.fieldloctpe, proxyout.fieldloc)
    c.pyapi.decref(ref_obj)
    c.pyapi.decref(fieldloc_obj)
    c.pyapi.decref(array_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(IdentitiesType)
def box(tpe, val, c):
    if tpe.bitwidth() == 32:
        Identities_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.Identities32))
    elif tpe.bitwidth() == 64:
        Identities_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.Identities64))
    else:
        assert False, "unrecognized bitwidth"
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    ref_obj = c.pyapi.from_native_value(util.RefType, proxyin.ref, c.env_manager)
    fieldloc_obj = c.pyapi.from_native_value(tpe.fieldloctpe, proxyin.fieldloc, c.env_manager)
    array_obj = c.pyapi.from_native_value(tpe.arraytpe, proxyin.array, c.env_manager)
    out = c.pyapi.call_function_objargs(Identities_obj, (ref_obj, fieldloc_obj, array_obj))
    c.pyapi.decref(Identities_obj)
    c.pyapi.decref(ref_obj)
    c.pyapi.decref(fieldloc_obj)
    c.pyapi.decref(array_obj)
    return out

@numba.extending.lower_builtin(len, IdentitiesType)
def lower_len(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    return numba.targets.arrayobj.array_len(context, builder, numba.types.intp(tpe.arraytpe), (proxyin.array,))

def lower_getitem_any(context, builder, identitiestpe, wheretpe, idval, whereval):
    proxyin = numba.cgutils.create_struct_proxy(identitiestpe)(context, builder, value=idval)

    if isinstance(wheretpe, numba.types.Integer):
        proxyslice = numba.cgutils.create_struct_proxy(numba.types.slice2_type)(context, builder)
        proxyslice.start = util.cast(context, builder, wheretpe, numba.intp, whereval)
        proxyslice.stop = builder.add(proxyslice.start, context.get_constant(numba.intp, 1))
        proxyslice.step = context.get_constant(numba.intp, 1)
        wheretpe = numba.types.slice2_type
        whereval = proxyslice._getvalue()

    proxyout = numba.cgutils.create_struct_proxy(identitiestpe)(context, builder)
    proxyout.ref = proxyin.ref
    proxyout.fieldloc = proxyin.fieldloc
    if isinstance(wheretpe, numba.types.BaseTuple):
        proxyout.array = numba.targets.arrayobj.getitem_array_tuple(context, builder, identitiestpe.arraytpe(identitiestpe.arraytpe, wheretpe), (proxyin.array, whereval))
    elif isinstance(wheretpe, numba.types.Array):
        proxyout.array = numba.targets.arrayobj.fancy_getitem_array(context, builder, identitiestpe.arraytpe(identitiestpe.arraytpe, wheretpe), (proxyin.array, whereval))
    else:
        proxyout.array = numba.targets.arrayobj.getitem_arraynd_intp(context, builder, identitiestpe.arraytpe(identitiestpe.arraytpe, wheretpe), (proxyin.array, whereval))

    return proxyout._getvalue()

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = IdentitiesType

    def generic_resolve(self, tpe, attr):
        if attr == "width":
            return numba.int64

@numba.extending.lower_getattr(IdentitiesType, "width")
def lower_content(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    array_proxy = numba.cgutils.create_struct_proxy(tpe.arraytpe)(context, builder, value=proxyin.array)
    out = builder.extract_value(array_proxy.shape, 1)
    return util.cast(context, builder, numba.intp, numba.int64, out)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test():
    out = awkward1.FillableArray()
    
    out.beginrecord()
    if True:
        out.field("x");       out.integer(3)
        out.field("extreme"); out.beginrecord()
        if True:
            out.field("pt");     out.real(3.3)
            out.field("charge"); out.integer(-1)
            out.field("iso");    out.integer(100)
        out.endrecord()
    out.endrecord()

    out.beginrecord()
    if True:
        out.field("x"); out.integer(3)
    out.endrecord()

    ss = out.snapshot()
    assert awkward1.tolist(ss) == [{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}}, {"x": 3, "extreme": None}]

    assert awkward1.tolist(awkward1.Array([{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}}, {"x": 3}])) == [{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}}, {"x": 3, "extreme": None}]

    assert awkward1.tolist(awkward1.Array([{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}}, {"x": 3, "what": 3}])) == [{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}, "what": None}, {"x": 3, "extreme": None, "what": 3}]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import numpy

import awkward1

def test():
    data = numpy.array([0, 2, 2, 3], dtype="i8")
    offsets = awkward1.layout.Index64(data)
    assert numpy.asarray(offsets).tolist() == [0, 2, 2, 3]
    assert offsets[0] == 0
    assert offsets[1] == 2
    assert offsets[2] == 2
    assert offsets[3] == 3
    data[2] = 999
    assert offsets[2] == 999

    data = numpy.array([0, 2, 2, 3], dtype="i4")
    offsets = awkward1.layout.Index32(data)
    assert numpy.asarray(offsets).tolist() == [0, 2, 2, 3]
    assert offsets[0] == 0
    assert offsets[1] == 2
    assert offsets[2] == 2
    assert offsets[3] == 3
    data[2] = 999
    assert offsets[2] == 999

    content = awkward1.layout.NumpyArray(numpy.arange(12).reshape(3, 4))
    assert numpy.asarray(content).tolist() == [[0,  1,  2,  3],
                                               [4,  5,  6,  7],
                                               [8,  9, 10, 11]]
    assert numpy.asarray(content[0]).tolist() == [0,  1,  2,  3]
    assert numpy.asarray(content[1]).tolist() == [4,  5,  6,  7]
    assert numpy.asarray(content[2]).tolist() == [8,  9, 10, 11]
    assert [content[i][j] for i in range(3) for j in range(4)] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

    data = numpy.array([0, 2, 2, 3], dtype="i4")
    offsets = awkward1.layout.Index32(data)
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert numpy.asarray(array[0]).tolist() == [[0, 1, 2, 3], [4, 5, 6, 7]]
    assert numpy.asarray(array[1]).tolist() == []
    assert numpy.asarray(array[2]).tolist() == [[8, 9, 10, 11]]
    assert numpy.asarray(array[1:3][0]).tolist() == []
    assert numpy.asarray(array[1:3][1]).tolist() == [[8, 9, 10, 11]]
    assert numpy.asarray(array[2:3][0]).tolist() == [[8, 9, 10, 11]]

def test_len():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], dtype="i4"))
    content = awkward1.layout.NumpyArray(numpy.arange(12).reshape(4, 3))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert len(content) == 4
    assert len(array) == 3

def test_members():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], dtype="i4"))
    content = awkward1.layout.NumpyArray(numpy.arange(12).reshape(3, 4))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert numpy.asarray(array.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(array.content).tolist() == [[0,  1,  2,  3], [4,  5,  6,  7], [8,  9, 10, 11]]
    array2 = awkward1.layout.ListOffsetArray32(offsets, array)
    assert numpy.asarray(array2.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(array2.content.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(array2.content.content).tolist() == [[0,  1,  2,  3], [4,  5,  6,  7], [8,  9, 10, 11]]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_count():
    content = awkward1.layout.NumpyArray(numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=numpy.int64))
    starts = awkward1.layout.Index64(numpy.array([4, 999, 0, 0, 1, 7], dtype=numpy.int64))
    stops = awkward1.layout.Index64(numpy.array([7, 999, 1, 4, 5, 10], dtype=numpy.int64))
    array1 = awkward1.layout.ListArray64(starts, stops, content)
    assert awkward1.tolist(array1) == [[4, 5, 6], [], [0], [0, 1, 2, 3], [1, 2, 3, 4], [7, 8, 9]]
    assert awkward1.tolist(array1.count()) == [3, 0, 1, 4, 4, 3]

    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10], dtype=numpy.int64))
    array2 = awkward1.layout.ListOffsetArray64(offsets, content)
    assert awkward1.tolist(array2) == [[0, 1, 2], [], [3, 4], [5], [6, 7, 8, 9]]
    assert awkward1.tolist(array2.count()) == [3, 0, 2, 1, 4]

    array3 = awkward1.layout.RegularArray(content, 5)
    assert awkward1.tolist(array3) == [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    assert awkward1.tolist(array3.count()) == [5, 5]

    content2 = awkward1.layout.NumpyArray(numpy.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=numpy.int64))
    assert awkward1.tolist(content2) == [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    assert awkward1.tolist(content2.count()) == [5, 5]

    index1 = awkward1.layout.Index64(numpy.array([2, 4, 0, 0, 1, 3], dtype=numpy.int64))
    array4 = awkward1.layout.IndexedArray64(index1, array2)
    assert awkward1.tolist(array4) == [[3, 4], [6, 7, 8, 9], [0, 1, 2], [0, 1, 2], [], [5]]
    assert awkward1.tolist(array4.count()) == [2, 4, 3, 3, 0, 1]

def test_indexedarray():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    index1 = awkward1.layout.Index64(numpy.array([2, 3, 3, 0, 4, 8], dtype=numpy.int64))
    indexedarray1 = awkward1.layout.IndexedArray64(index1, content)
    index2 = awkward1.layout.Index64(numpy.array([2, 3, 3, -1, -1, 8], dtype=numpy.int64))
    indexedarray2 = awkward1.layout.IndexedOptionArray64(index2, content)
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6], dtype=numpy.int64))
    listoffsetarray1 = awkward1.layout.ListOffsetArray64(offsets1, indexedarray1)
    listoffsetarray2 = awkward1.layout.ListOffsetArray64(offsets1, indexedarray2)
    offsets3 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10], dtype=numpy.int64))
    listoffsetarray3 = awkward1.layout.ListOffsetArray64(offsets3, content)
    index3 = awkward1.layout.Index64(numpy.array([2, 0, 1, 3, 3, 4], dtype=numpy.int64))
    indexedarray3 = awkward1.layout.IndexedArray64(index3, listoffsetarray3)
    index4 = awkward1.layout.Index64(numpy.array([2, -1, -1, 3, 3, 4], dtype=numpy.int64))
    indexedarray4 = awkward1.layout.IndexedOptionArray64(index4, listoffsetarray3)

    assert awkward1.tolist(indexedarray1) == [2.2, 3.3, 3.3, 0.0, 4.4, 8.8]
    assert awkward1.tolist(indexedarray2) == [2.2, 3.3, 3.3, None, None, 8.8]
    assert awkward1.tolist(listoffsetarray1) == [[2.2, 3.3, 3.3], [], [0.0, 4.4], [8.8]]
    assert awkward1.tolist(listoffsetarray2) == [[2.2, 3.3, 3.3], [], [None, None], [8.8]]
    assert awkward1.tolist(listoffsetarray3) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.tolist(indexedarray3) == [[3.3, 4.4], [0.0, 1.1, 2.2], [], [5.5], [5.5], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.tolist(indexedarray4) == [[3.3, 4.4], None, None, [5.5], [5.5], [6.6, 7.7, 8.8, 9.9]]

    with pytest.raises(ValueError) as err:
        indexedarray1.flatten()
    assert str(err.value) == "NumpyArray cannot be flattened because it has 1 dimensions"

    with pytest.raises(ValueError) as err:
        indexedarray2.flatten()
    assert str(err.value) == "NumpyArray cannot be flattened because it has 1 dimensions"

    assert awkward1.tolist(indexedarray3.flatten()) == [3.3, 4.4, 0.0, 1.1, 2.2, 5.5, 5.5, 6.6, 7.7, 8.8, 9.9]

    assert awkward1.tolist(indexedarray4.flatten()) == [3.3, 4.4, 5.5, 5.5, 6.6, 7.7, 8.8, 9.9]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1

def test_iterator():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert list(content) == [1.1, 2.2, 3.3]
    assert [numpy.asarray(x).tolist() for x in array] == [[1.1, 2.2], [], [3.3]]

def test_refcount():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)

    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)

    iter1 = iter(content)
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)
    x1 = next(iter1)
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)

    iter2 = iter(array)
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)
    x2 = next(iter2)
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)

    del iter1
    del x1
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)

    del iter2
    del x2
    assert (sys.getrefcount(content), sys.getrefcount(array)) == (2, 2)

def test_numba_numpyarray():
    array = awkward1.layout.NumpyArray(numpy.arange(12))
    @numba.njit
    def f1(q):
        out = 0
        for x in q:
            out += x
        return out
    assert f1(array) == 66

    array = awkward1.layout.NumpyArray(numpy.arange(12).reshape(3, 4))
    @numba.njit
    def f2(q):
        out = 0
        for x in q:
            for y in x:
                out += y
        return out
    assert f2(array) == 66

def test_numba_listoffsetarray():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)

    @numba.njit
    def f1(q):
        out = 0.0
        for x in q:
            for y in x:
                out += y
        return out
    assert f1(array) == 6.6

    @numba.njit
    def f2(q):
        for x in q:
            return x
    assert numpy.asarray(f2(array)).tolist() == [1.1, 2.2]

def test_tolist():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert awkward1.tolist(array) == [[1.1, 2.2], [], [3.3]]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

def test_basic():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5]))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
    recordarray = awkward1.layout.RecordArray(0)
    recordarray.append(content1, "one")
    recordarray.append(listoffsetarray, "two")
    recordarray.append(content2)
    recordarray.append(content1, "wonky")
    assert awkward1.tolist(recordarray.field(0)) == [1, 2, 3, 4, 5]
    assert awkward1.tolist(recordarray.field("two")) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(recordarray.field("wonky")) == [1, 2, 3, 4, 5]

    str(recordarray)
    assert awkward1.tojson(recordarray) == '[{"one":1,"two":[1.1,2.2,3.3],"2":1.1,"wonky":1},{"one":2,"two":[],"2":2.2,"wonky":2},{"one":3,"two":[4.4,5.5],"2":3.3,"wonky":3},{"one":4,"two":[6.6],"2":4.4,"wonky":4},{"one":5,"two":[7.7,8.8,9.9],"2":5.5,"wonky":5}]'

    assert len(recordarray) == 5
    assert recordarray.key(0) == "one"
    assert recordarray.key(1) == "two"
    assert recordarray.key(2) == "2"
    assert recordarray.key(3) == "wonky"
    assert recordarray.fieldindex("wonky") == 3
    assert recordarray.fieldindex("one") == 0
    assert recordarray.fieldindex("0") == 0
    assert recordarray.fieldindex("two") == 1
    assert recordarray.fieldindex("1") == 1
    assert recordarray.fieldindex("2") == 2
    assert recordarray.haskey("wonky")
    assert recordarray.haskey("one")
    assert recordarray.haskey("0")
    assert recordarray.haskey("two")
    assert recordarray.haskey("1")
    assert recordarray.haskey("2")

    assert recordarray.keys() == ["one", "two", "2", "wonky"]
    assert [awkward1.tolist(x) for x in recordarray.fields()] == [[1, 2, 3, 4, 5], [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], [1, 2, 3, 4, 5]]
    pairs = recordarray.fielditems()
    assert pairs[0][0] == "one"
    assert pairs[1][0] == "two"
    assert pairs[2][0] == "2"
    assert pairs[3][0] == "wonky"
    assert awkward1.tolist(pairs[0][1]) == [1, 2, 3, 4, 5]
    assert awkward1.tolist(pairs[1][1]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(pairs[2][1]) == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]
    assert awkward1.tolist(pairs[3][1]) == [1, 2, 3, 4, 5]

    assert awkward1.tojson(recordarray.astuple) == '[{"0":1,"1":[1.1,2.2,3.3],"2":1.1,"3":1},{"0":2,"1":[],"2":2.2,"3":2},{"0":3,"1":[4.4,5.5],"2":3.3,"3":3},{"0":4,"1":[6.6],"2":4.4,"3":4},{"0":5,"1":[7.7,8.8,9.9],"2":5.5,"3":5}]'

def test_scalar_record():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5]))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
    recordarray = awkward1.layout.RecordArray(0)
    recordarray.append(content1, "one")
    recordarray.append(listoffsetarray, "two")

    str(recordarray)
    str(recordarray[2])
    assert awkward1.tojson(recordarray[2]) == '{"one":3,"two":[4.4,5.5]}'

    assert recordarray[2].keys() == ["one", "two"]
    assert [awkward1.tolist(x) for x in recordarray[2].fields()] == [3, [4.4, 5.5]]
    pairs = recordarray[2].fielditems()
    assert pairs[0][0] == "one"
    assert pairs[1][0] == "two"
    assert pairs[0][1] == 3
    assert awkward1.tolist(pairs[1][1]) == [4.4, 5.5]
    assert awkward1.tolist(recordarray[2]) == {"one": 3, "two": [4.4, 5.5]}

    assert awkward1.tolist(awkward1.layout.Record(recordarray, 2)) == {"one": 3, "two": [4.4, 5.5]}

def test_type():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=numpy.float64))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
    recordarray = awkward1.layout.RecordArray(0, True)
    recordarray.append(content1)
    recordarray.append(listoffsetarray)
    assert str(awkward1.typeof(recordarray)) == '(int64, var * float64)'

    assert awkward1.typeof(recordarray) == awkward1.layout.RecordType((
        awkward1.layout.PrimitiveType("int64"),
        awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"))))
    assert awkward1.typeof(recordarray[2]) == awkward1.layout.RecordType(
        (awkward1.layout.PrimitiveType("int64"),
        awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"))))

    recordarray = awkward1.layout.RecordArray(0, True)
    recordarray.append(content1, "one")
    recordarray.append(listoffsetarray, "two")
    assert str(awkward1.typeof(recordarray)) in ('{"one": int64, "two": var * float64}', '{"two": var * float64, "one": int64}')

    assert str(awkward1.layout.RecordType(
        (awkward1.layout.PrimitiveType("int32"),
        awkward1.layout.PrimitiveType("float64")))) == '(int32, float64)'

    assert str(awkward1.layout.RecordType(
        {"one": awkward1.layout.PrimitiveType("int32"),
        "two": awkward1.layout.PrimitiveType("float64")})) in ('{"one": int32, "two": float64}', '{"two": float64, "one": int32}')

    assert awkward1.typeof(recordarray) == awkward1.layout.RecordType({
        "one": awkward1.layout.PrimitiveType("int64"),
        "two": awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"))})
    assert awkward1.typeof(recordarray[2]) == awkward1.layout.RecordType({
        "one": awkward1.layout.PrimitiveType("int64"),
        "two": awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"))})

def test_getitem():
    assert str(awkward1.layout.Slice((1, 2, [3], "four", ["five", "six"], slice(7, 8, 9)))) == '[array([1]), array([2]), array([3]), "four", ["five", "six"], 7:8:9]'

    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=numpy.float64))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
    recordarray = awkward1.layout.RecordArray([content1, listoffsetarray, content2])
    assert recordarray.istuple

    assert awkward1.tolist(recordarray["2"]) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert awkward1.tolist(recordarray[["0", "1"]]) == [(1, [1.1, 2.2, 3.3]), (2, []), (3, [4.4, 5.5]), (4, [6.6]), (5, [7.7, 8.8, 9.9])]
    assert awkward1.tolist(recordarray[["1", "0"]]) == [([1.1, 2.2, 3.3], 1), ([], 2), ([4.4, 5.5], 3), ([6.6], 4), ([7.7, 8.8, 9.9], 5)]
    assert awkward1.tolist(recordarray[1:-1]) == [(2, [], 2.2), (3, [4.4, 5.5], 3.3), (4, [6.6], 4.4)]
    assert awkward1.tolist(recordarray[2]) == (3, [4.4, 5.5], 3.3)
    assert awkward1.tolist(recordarray[2]["1"]) == [4.4, 5.5]
    assert awkward1.tolist(recordarray[2][["0", "1"]]) == (3, [4.4, 5.5])
    assert awkward1.tolist(recordarray[2][["1", "0"]]) == ([4.4, 5.5], 3)

    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray, "three": content2})
    assert not recordarray.istuple

    assert awkward1.tolist(recordarray["three"]) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert awkward1.tolist(recordarray[["one", "two"]]) == [{"one": 1, "two": [1.1, 2.2, 3.3]}, {"one": 2, "two": []}, {"one": 3, "two": [4.4, 5.5]}, {"one": 4, "two": [6.6]}, {"one": 5, "two": [7.7, 8.8, 9.9]}]
    assert awkward1.tolist(recordarray[["two", "one"]]) == [{"one": 1, "two": [1.1, 2.2, 3.3]}, {"one": 2, "two": []}, {"one": 3, "two": [4.4, 5.5]}, {"one": 4, "two": [6.6]}, {"one": 5, "two": [7.7, 8.8, 9.9]}]
    assert awkward1.tolist(recordarray[1:-1]) == [{"one": 2, "two": [], "three": 2.2}, {"one": 3, "two": [4.4, 5.5], "three": 3.3}, {"one": 4, "two": [6.6], "three": 4.4}]
    assert awkward1.tolist(recordarray[2]) == {"one": 3, "two": [4.4, 5.5], "three": 3.3}
    assert awkward1.tolist(recordarray[2]["two"]) == [4.4, 5.5]
    assert awkward1.tolist(recordarray[2][["one", "two"]]) == {"one": 3, "two": [4.4, 5.5]}
    assert awkward1.tolist(recordarray[2][["two", "one"]]) == {"one": 3, "two": [4.4, 5.5]}

def test_getitem_other_types():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=numpy.float64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray1 = awkward1.layout.ListOffsetArray64(offsets1, content2)
    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray1, "three": content2})

    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5]))
    listoffsetarray2 = awkward1.layout.ListOffsetArray64(offsets2, recordarray)
    assert awkward1.tolist(listoffsetarray2["one"]) == [[1, 2, 3], [], [4, 5]]
    assert awkward1.tolist(listoffsetarray2["two"]) == [[[1.1, 2.2, 3.3], [], [4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]]
    assert awkward1.tolist(listoffsetarray2["three"]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(listoffsetarray2[["two", "three"]]) == [[{"two": [1.1, 2.2, 3.3], "three": 1.1}, {"two": [], "three": 2.2}, {"two": [4.4, 5.5], "three": 3.3}], [], [{"two": [6.6], "three": 4.4}, {"two": [7.7, 8.8, 9.9], "three": 5.5}]]

    starts2 = awkward1.layout.Index64(numpy.array([0, 3, 3]))
    stops2 = awkward1.layout.Index64(numpy.array([3, 3, 5]))
    listarray2 = awkward1.layout.ListArray64(starts2, stops2, recordarray)
    assert awkward1.tolist(listarray2["one"]) == [[1, 2, 3], [], [4, 5]]
    assert awkward1.tolist(listarray2["two"]) == [[[1.1, 2.2, 3.3], [], [4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]]
    assert awkward1.tolist(listarray2["three"]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(listarray2[["two", "three"]]) == [[{"two": [1.1, 2.2, 3.3], "three": 1.1}, {"two": [], "three": 2.2}, {"two": [4.4, 5.5], "three": 3.3}], [], [{"two": [6.6], "three": 4.4}, {"two": [7.7, 8.8, 9.9], "three": 5.5}]]

    regulararray2 = awkward1.layout.RegularArray(recordarray, 1)
    assert awkward1.tolist(regulararray2["one"]) == [[1], [2], [3], [4], [5]]
    assert awkward1.tolist(regulararray2["two"]) == [[[1.1, 2.2, 3.3]], [[]], [[4.4, 5.5]], [[6.6]], [[7.7, 8.8, 9.9]]]
    assert awkward1.tolist(regulararray2["three"]) == [[1.1], [2.2], [3.3], [4.4], [5.5]]
    assert awkward1.tolist(regulararray2[["two", "three"]]) == [[{"two": [1.1, 2.2, 3.3], "three": 1.1}], [{"two": [], "three": 2.2}], [{"two": [4.4, 5.5], "three": 3.3}], [{"two": [6.6], "three": 4.4}], [{"two": [7.7, 8.8, 9.9], "three": 5.5}]]

def test_getitem_next():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=numpy.float64))
    content3 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=numpy.float64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray1 = awkward1.layout.ListOffsetArray64(offsets1, content2)
    listoffsetarray3 = awkward1.layout.ListOffsetArray64(offsets1, content3)
    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray1, "three": content2, "four": listoffsetarray3})
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5]))
    listoffsetarray2 = awkward1.layout.ListOffsetArray64(offsets2, recordarray)

    assert awkward1.tolist(listoffsetarray2[2, "one"]) == [4, 5]
    assert awkward1.tolist(listoffsetarray2[2, "two"]) == [[6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(listoffsetarray2[2, "three"]) == [4.4, 5.5]
    assert awkward1.tolist(listoffsetarray2[2, ["two", "three"]]) == [{"two": [6.6], "three": 4.4}, {"two": [7.7, 8.8, 9.9], "three": 5.5}]

    assert awkward1.tolist(listoffsetarray2[2, 1]) == {"one": 5, "two": [7.7, 8.8, 9.9], "three": 5.5, "four": [7, 8, 9]}
    with pytest.raises(ValueError):
        listoffsetarray2[2, 1, 0]
    assert listoffsetarray2[2, 1, "one"] == 5
    assert awkward1.tolist(listoffsetarray2[2, 1, "two"]) == [7.7, 8.8, 9.9]
    assert listoffsetarray2[2, 1, "two", 1] == 8.8
    assert awkward1.tolist(listoffsetarray2[2, 1, ["two", "four"], 1]) == {"two": 8.8, "four": 8}
    assert awkward1.tolist(listoffsetarray2[2, 1, ["two", "four"], 1:]) == {"two": [8.8, 9.9], "four": [8, 9]}

def test_setidentities():
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5]))
    content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)

    recordarray = awkward1.layout.RecordArray([content1, listoffsetarray])
    recordarray.setidentities()

    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray})
    recordarray.setidentities()
    assert recordarray["one"].identities.fieldloc == [(0, "one")]
    assert recordarray["two"].identities.fieldloc == [(0, "two")]
    assert recordarray["one", 1] == 2
    assert recordarray[1, "one"] == 2
    assert recordarray["two", 2, 1] == 5.5
    assert recordarray[2, "two", 1] == 5.5

    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray})
    recordarray2 = awkward1.layout.RecordArray({"outer": recordarray})
    recordarray2.setidentities()
    assert recordarray2["outer"].identities.fieldloc == [(0, "outer")]
    assert recordarray2["outer", "one"].identities.fieldloc == [(0, "outer"), (0, "one")]
    assert recordarray2["outer", "two"].identities.fieldloc == [(0, "outer"), (0, "two")]
    assert recordarray2["outer", "one", 1] == 2
    assert recordarray2["outer", 1, "one"] == 2
    assert recordarray2[1, "outer", "one"] == 2
    assert recordarray2["outer", "two", 2, 1] == 5.5
    assert recordarray2["outer", 2, "two", 1] == 5.5
    assert recordarray2[2, "outer", "two", 1] == 5.5
    with pytest.raises(ValueError) as excinfo:
        recordarray2["outer", "two", 0, 99]
    assert str(excinfo.value) == 'in ListArray64 with identity [0, "outer", "two"] attempting to get 99, index out of range'
    assert recordarray2.identity == ()
    assert recordarray2[2].identity == (2,)
    assert recordarray2[2, "outer"].identity == (2, "outer")
    assert recordarray2[2, "outer", "two"].identity == (2, "outer", "two")

    recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray})
    recordarray2 = awkward1.layout.RecordArray({"outer": awkward1.layout.RegularArray(recordarray, 1)})
    recordarray2.setidentities()
    assert recordarray2["outer"].identities.fieldloc == [(0, "outer")]
    assert recordarray2["outer", 0, "one"].identities.fieldloc == [(0, "outer"), (1, "one")]
    assert recordarray2["outer", 0, "two"].identities.fieldloc == [(0, "outer"), (1, "two")]
    assert recordarray2["outer", "one", 0].identities.fieldloc == [(0, "outer"), (1, "one")]
    assert recordarray2["outer", "two", 0].identities.fieldloc == [(0, "outer"), (1, "two")]
    assert recordarray2["outer", "one", 1, 0] == 2
    assert recordarray2["outer", 1, "one", 0] == 2
    assert recordarray2["outer", 1, 0, "one"] == 2
    assert recordarray2[1, "outer", "one", 0] == 2
    assert recordarray2[1, "outer", 0, "one"] == 2
    assert recordarray2[1, 0, "outer", "one"] == 2

    with pytest.raises(ValueError) as excinfo:
        recordarray2["outer", 2, "two", 0, 99]
    assert str(excinfo.value) == 'in ListArray64 with identity [2, "outer", 0, "two"] attempting to get 99, index out of range'
    assert recordarray2.identity == ()
    assert recordarray2[2].identity == (2,)
    assert recordarray2[2, "outer"].identity == (2, "outer")
    assert recordarray2[2, "outer", 0].identity == (2, "outer", 0)
    assert recordarray2[2, "outer", 0, "two"].identity == (2, "outer", 0, "two")

def test_fillable_tuple():
    fillable = awkward1.layout.FillableArray()
    assert str(fillable.type) == 'unknown'
    assert awkward1.tolist(fillable.snapshot()) == []

    fillable.begintuple(0)
    fillable.endtuple()

    fillable.begintuple(0)
    fillable.endtuple()

    fillable.begintuple(0)
    fillable.endtuple()

    assert str(fillable.type) == '()'
    assert awkward1.tolist(fillable.snapshot()) == [(), (), ()]

    fillable = awkward1.layout.FillableArray()

    fillable.begintuple(3)
    fillable.index(0)
    fillable.boolean(True)
    fillable.index(1)
    fillable.beginlist()
    fillable.integer(1)
    fillable.endlist()
    fillable.index(2)
    fillable.real(1.1)
    fillable.endtuple()

    fillable.begintuple(3)
    fillable.index(1)
    fillable.beginlist()
    fillable.integer(2)
    fillable.integer(2)
    fillable.endlist()
    fillable.index(2)
    fillable.real(2.2)
    fillable.index(0)
    fillable.boolean(False)
    fillable.endtuple()

    fillable.begintuple(3)
    fillable.index(2)
    fillable.real(3.3)
    fillable.index(1)
    fillable.beginlist()
    fillable.integer(3)
    fillable.integer(3)
    fillable.integer(3)
    fillable.endlist()
    fillable.index(0)
    fillable.boolean(True)
    fillable.endtuple()

    assert str(fillable.type) == '(bool, var * int64, float64)'
    assert awkward1.tolist(fillable.snapshot()) == [(True, [1], 1.1), (False, [2, 2], 2.2), (True, [3, 3, 3], 3.3)]

def test_fillable_record():
    fillable = awkward1.layout.FillableArray()
    assert str(fillable.type) == 'unknown'
    assert awkward1.tolist(fillable.snapshot()) == []

    fillable.beginrecord()
    fillable.endrecord()

    fillable.beginrecord()
    fillable.endrecord()

    fillable.beginrecord()
    fillable.endrecord()

    assert str(fillable.type) == '{}'
    assert awkward1.tolist(fillable.snapshot()) == [{}, {}, {}]

    fillable = awkward1.layout.FillableArray()

    fillable.beginrecord()
    fillable.field("one")
    fillable.integer(1)
    fillable.field("two")
    fillable.real(1.1)
    fillable.endrecord()

    fillable.beginrecord()
    fillable.field("two")
    fillable.real(2.2)
    fillable.field("one")
    fillable.integer(2)
    fillable.endrecord()

    fillable.beginrecord()
    fillable.field("one")
    fillable.integer(3)
    fillable.field("two")
    fillable.real(3.3)
    fillable.endrecord()

    assert str(fillable.type) == '{"one": int64, "two": float64}'
    assert awkward1.tolist(fillable.snapshot()) == [{"one": 1, "two": 1.1}, {"one": 2, "two": 2.2}, {"one": 3, "two": 3.3}]

def test_fromiter():
    dataset = [
        [(1, 1.1), (2, 2.2), (3, 3.3)],
        [(1, [1.1, 2.2, 3.3]), (2, []), (3, [4.4, 5.5])],
        [[(1, 1.1), (2, 2.2), (3, 3.3)], [], [(4, 4.4), (5, 5.5)]],
        [((1, 1), 1.1), ((2, 2), 2.2), ((3, 3), 3.3)],
        [({"x": 1, "y": 1}, 1.1), ({"x": 2, "y": 2}, 2.2), ({"x": 3, "y": 3}, 3.3)],
        [{"one": 1, "two": 1.1}, {"one": 2, "two": 2.2}, {"one": 3, "two": 3.3}],
        [{"one": 1, "two": [1.1, 2.2, 3.3]}, {"one": 2, "two": []}, {"one": 3, "two": [4.4, 5.5]}],
        [[{"one": 1, "two": 1.1}, {"one": 2, "two": 2.2}, {"one": 3, "two": 3.3}], [], [{"one": 4, "two": 4.4}, {"one": 5, "two": 5.5}]],
        [{"one": {"x": 1, "y": 1}, "two": 1.1}, {"one": {"x": 2, "y": 2}, "two": 2.2}, {"one": {"x": 3, "y": 3}, "two": 3.3}],
        [{"one": (1, 1), "two": 1.1}, {"one": (2, 2), "two": 2.2}, {"one": (3, 3), "two": 3.3}],
    ]
    for datum in dataset:
        assert awkward1.tolist(awkward1.fromiter(datum)) == datum

def test_json():
    dataset = [
        '[{"one":1,"two":1.1},{"one":2,"two":2.2},{"one":3,"two":3.3}]',
        '[{"one":1,"two":[1.1,2.2,3.3]},{"one":2,"two":[]},{"one":3,"two":[4.4,5.5]}]',
        '[[{"one":1,"two":1.1},{"one":2,"two":2.2},{"one":3,"two":3.3}],[],[{"one":4,"two":4.4},{"one":5,"two":5.5}]]',
        '[{"one":{"x":1,"y":1},"two":1.1},{"one":{"x":2,"y":2},"two":2.2},{"one":{"x":3,"y":3},"two":3.3}]',
    ]
    for datum in dataset:
        assert awkward1.tojson(awkward1.fromjson(datum)) == datum



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools
import pickle

import pytest
import numpy

numba = pytest.importorskip("numba")

import awkward1

if sys.version_info[0] < 3:
    pytest.skip("pybind11 pickle, and hence numba serialization with types, only works in Python 3", allow_module_level=True)

def test_pickle():
    t = awkward1.layout.UnknownType(); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.PrimitiveType("int32"); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.PrimitiveType("float64"); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.utf8; assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.string; assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.ArrayType(awkward1.layout.PrimitiveType("int32"), 100); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.ListType(awkward1.layout.PrimitiveType("int32")); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.RegularType(awkward1.layout.PrimitiveType("int32"), 5); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.OptionType(awkward1.layout.PrimitiveType("int32")); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.UnionType((awkward1.layout.PrimitiveType("int32"), awkward1.layout.PrimitiveType("float64"))); assert pickle.loads(pickle.dumps(t)) == t
    t = awkward1.layout.RecordType({"one": awkward1.layout.PrimitiveType("int32"), "two": awkward1.layout.PrimitiveType("float64")}); assert pickle.loads(pickle.dumps(t)) == t

def test_boxing():
    @numba.njit
    def f1(q):
        return 3.14

    @numba.njit
    def f2(q):
        return q

    t = awkward1.layout.UnknownType()
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.PrimitiveType("int32")
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.PrimitiveType("float64")
    f1(t)
    assert f2(t) == t

    t = awkward1.utf8
    f1(t)
    assert f2(t) == t

    t = awkward1.string
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.PrimitiveType("float64")
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.ListType(awkward1.layout.ListType(awkward1.layout.PrimitiveType("int32")))
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.RegularType(awkward1.layout.ListType(awkward1.layout.PrimitiveType("int32")), 5)
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.OptionType(awkward1.layout.ListType(awkward1.layout.PrimitiveType("int32")))
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.UnionType((awkward1.layout.PrimitiveType("int32"), awkward1.layout.PrimitiveType("float64")))
    f1(t)
    assert f2(t) == t

    t = awkward1.layout.RecordType({"one": awkward1.layout.PrimitiveType("int32"), "two": awkward1.layout.PrimitiveType("float64")})
    f1(t)
    assert f2(t) == t

class D(awkward1.highlevel.Array):
    pass

def test_numpyarray():
    dint64 = awkward1.layout.PrimitiveType("int64", {"__class__": "D", "__typestr__": "D[int64]"})
    array1 = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64).reshape(2, 3, 5)).astype(awkward1.layout.RegularType(awkward1.layout.RegularType(dint64, 5), 3))

    @numba.njit
    def f1(q):
        return q

    array2 = f1(array1)

    assert repr(array2.type) == "3 * 5 * D[int64]"
    assert repr(array2[0].type) == "5 * D[int64]"
    assert repr(array2[0, 0].type) == "D[int64]"
    assert array2[-1, -1, -1] == 29

def test_regulararray():
    dregint64 = awkward1.layout.RegularType(awkward1.layout.PrimitiveType("int64"), 5, {"__class__": "D", "__typestr__": "D[5 * int64]"})
    array1 = awkward1.layout.RegularArray(awkward1.layout.NumpyArray(numpy.arange(10, dtype=numpy.int64)), 5).astype(dregint64)

    @numba.njit
    def f1(q):
        return q

    array2 = f1(array1)

    assert repr(array2.type) == "D[5 * int64]"

def test_listoffsetarray():
    dvarint64 = awkward1.layout.ListType(awkward1.layout.PrimitiveType("int64"), {"__class__": "D", "__typestr__": "D[var * int64]"})
    array1 = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 5], dtype=numpy.int64)), awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))).astype(dvarint64)

    @numba.njit
    def f1(q):
        return q

    array2 = f1(array1)

    assert repr(array2.type) == "D[var * int64]"

def test_listarray():
    dvarint64 = awkward1.layout.ListType(awkward1.layout.PrimitiveType("int64"), {"__class__": "D", "__typestr__": "D[var * int64]"})
    array1 = awkward1.layout.ListArray64(awkward1.layout.Index64(numpy.array([0, 3, 3], dtype=numpy.int64)), awkward1.layout.Index64(numpy.array([3, 3, 5], dtype=numpy.int64)), awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5], dtype=numpy.int64))).astype(dvarint64)

    @numba.njit
    def f1(q):
        return q

    array2 = f1(array1)

    assert repr(array2.type) == "D[var * int64]"

def test_recordarray():
    dvarrec = awkward1.layout.RecordType({"one": awkward1.layout.PrimitiveType("int64"), "two": awkward1.layout.PrimitiveType("float64")}, {"__class__": "D", "__typestr__": "D[{\"one\": int64, \"two\": float64}]"})
    array1 = awkward1.Array([{"one": 1, "two": 1.1}, {"one": 2, "two": 2.2}, {"one": 3, "two": 3.3}]).layout.astype(dvarrec)

    @numba.njit
    def f1(q):
        return q

    array2 = f1(array1)

    assert repr(array2.type) in ('D[{"one": int64, "two": float64}]', ' D[{"two": float64, "one": int64}]')
    assert repr(array2[0].type) in ('D[{"one": int64, "two": float64}]', 'D[{"two": float64, "one": int64}]')



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import numba

import awkward1.layout
from .._numba import libawkward, util

@numba.extending.typeof_impl.register(awkward1.layout.FillableArray)
def typeof(val, c):
    return FillableArrayType()

class FillableArrayType(numba.types.Type):
    def __init__(self):
        super(FillableArrayType, self).__init__(name="ak::FillableArrayType")

@numba.typing.templates.infer_global(len)
class type_len(numba.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            arraytpe, = args
            if isinstance(arraytpe, FillableArrayType):
                return numba.typing.templates.signature(numba.types.intp, arraytpe)

@numba.datamodel.registry.register_default(FillableArrayType)
class FillableArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("rawptr", numba.types.voidptr),
                   ("pyptr", numba.types.pyobject)]
        super(FillableArrayModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(FillableArrayType)
def unbox(tpe, obj, c):
    rawptr_obj = c.pyapi.object_getattr_string(obj, "_ptr")
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    proxyout.rawptr = c.pyapi.long_as_voidptr(rawptr_obj)
    proxyout.pyptr = obj
    c.pyapi.decref(rawptr_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(FillableArrayType)
def box(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    c.pyapi.incref(proxyin.pyptr)
    return proxyin.pyptr

def call(context, builder, fcn, args):
    fcntpe = context.get_function_pointer_type(fcn.numbatpe)
    fcnval = context.add_dynamic_addr(builder, fcn.numbatpe.get_pointer(fcn), info=fcn.name)
    fcnptr = builder.bitcast(fcnval, fcntpe)
    err = context.call_function_pointer(builder, fcnptr, args)
    with builder.if_then(builder.icmp_unsigned("!=", err, context.get_constant(numba.uint8, 0)), likely=False):
        context.call_conv.return_user_exc(builder, ValueError, (fcn.name + " failed",))

@numba.extending.lower_builtin(len, FillableArrayType)
def lower_len(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    result = numba.cgutils.alloca_once(builder, context.get_value_type(numba.int64))
    call(context, builder, libawkward.FillableArray_length, (proxyin.rawptr, result))
    return util.cast(context, builder, numba.int64, numba.intp, builder.load(result))

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = FillableArrayType

    @numba.typing.templates.bound_function("clear")
    def resolve_clear(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.clear")

    @numba.typing.templates.bound_function("null")
    def resolve_null(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.null")

    @numba.typing.templates.bound_function("boolean")
    def resolve_boolean(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.Boolean):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number or types of arguments for FillableArray.boolean")

    @numba.typing.templates.bound_function("integer")
    def resolve_integer(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.Integer):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number or types of arguments for FillableArray.integer")

    @numba.typing.templates.bound_function("real")
    def resolve_real(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], (numba.types.Integer, numba.types.Float)):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number or types of arguments for FillableArray.real")

    @numba.typing.templates.bound_function("beginlist")
    def resolve_beginlist(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.beginlist")

    @numba.typing.templates.bound_function("endlist")
    def resolve_endlist(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.endlist")

    @numba.typing.templates.bound_function("begintuple")
    def resolve_begintuple(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.Integer):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number of arguments for FillableArray.begintuple")

    @numba.typing.templates.bound_function("index")
    def resolve_index(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.Integer):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number of arguments for FillableArray.index")

    @numba.typing.templates.bound_function("endtuple")
    def resolve_endtuple(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.endtuple")

    @numba.typing.templates.bound_function("beginrecord")
    def resolve_beginrecord(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        elif len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.StringLiteral):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number of arguments for FillableArray.beginrecord")

    @numba.typing.templates.bound_function("field")
    def resolve_field(self, arraytpe, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], numba.types.StringLiteral):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number of arguments for FillableArray.field")

    @numba.typing.templates.bound_function("endrecord")
    def resolve_endrecord(self, arraytpe, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for FillableArray.endrecord")

@numba.extending.lower_builtin("clear", FillableArrayType)
def lower_clear(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_clear, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("null", FillableArrayType)
def lower_null(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_null, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("boolean", FillableArrayType, numba.types.Boolean)
def lower_integer(context, builder, sig, args):
    tpe, xtpe = sig.args
    val, xval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    x = builder.zext(xval, context.get_value_type(numba.uint8))
    call(context, builder, libawkward.FillableArray_boolean, (proxyin.rawptr, x))
    return context.get_dummy_value()

@numba.extending.lower_builtin("integer", FillableArrayType, numba.types.Integer)
def lower_integer(context, builder, sig, args):
    tpe, xtpe = sig.args
    val, xval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    x = util.cast(context, builder, xtpe, numba.int64, xval)
    call(context, builder, libawkward.FillableArray_integer, (proxyin.rawptr, x))
    return context.get_dummy_value()

@numba.extending.lower_builtin("real", FillableArrayType, numba.types.Integer)
@numba.extending.lower_builtin("real", FillableArrayType, numba.types.Float)
def lower_real(context, builder, sig, args):
    tpe, xtpe = sig.args
    val, xval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if isinstance(xtpe, numba.types.Integer) and xtpe.signed:
        x = builder.sitofp(xval, context.get_value_type(numba.types.float64))
    elif isinstance(xtpe, numba.types.Integer):
        x = builder.uitofp(xval, context.get_value_type(numba.types.float64))
    elif xtpe.bitwidth < 64:
        x = builder.fpext(xval, context.get_value_type(numba.types.float64))
    elif xtpe.bitwidth > 64:
        x = builder.fptrunc(xval, context.get_value_type(numba.types.float64))
    else:
        x = xval
    call(context, builder, libawkward.FillableArray_real, (proxyin.rawptr, x))
    return context.get_dummy_value()

@numba.extending.lower_builtin("beginlist", FillableArrayType)
def lower_beginlist(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_beginlist, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("endlist", FillableArrayType)
def lower_endlist(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_endlist, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("begintuple", FillableArrayType, numba.types.Integer)
def lower_begintuple(context, builder, sig, args):
    tpe, numfieldstpe = sig.args
    val, numfieldsval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    numfields = util.cast(context, builder, numfieldstpe, numba.int64, numfieldsval)
    call(context, builder, libawkward.FillableArray_begintuple, (proxyin.rawptr, numfields))
    return context.get_dummy_value()

@numba.extending.lower_builtin("index", FillableArrayType, numba.types.Integer)
def lower_index(context, builder, sig, args):
    tpe, indextpe = sig.args
    val, indexval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    index = util.cast(context, builder, indextpe, numba.int64, indexval)
    call(context, builder, libawkward.FillableArray_index, (proxyin.rawptr, index))
    return context.get_dummy_value()

@numba.extending.lower_builtin("endtuple", FillableArrayType)
def lower_endtuple(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_endtuple, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("beginrecord", FillableArrayType)
def lower_beginrecord(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_beginrecord, (proxyin.rawptr,))
    return context.get_dummy_value()

@numba.extending.lower_builtin("beginrecord", FillableArrayType, numba.types.StringLiteral)
def lower_beginrecord(context, builder, sig, args):
    tpe, nametpe = sig.args
    val, nameval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    name = util.globalstring(context, builder, nametpe.literal_value)
    call(context, builder, libawkward.FillableArray_beginrecord_fast, (proxyin.rawptr, name))
    return context.get_dummy_value()

@numba.extending.lower_builtin("field", FillableArrayType, numba.types.StringLiteral)
def lower_field(context, builder, sig, args):
    tpe, keytpe = sig.args
    val, keyval = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    key = util.globalstring(context, builder, keytpe.literal_value)
    call(context, builder, libawkward.FillableArray_field_fast, (proxyin.rawptr, key))
    return context.get_dummy_value()

@numba.extending.lower_builtin("endrecord", FillableArrayType)
def lower_endrecord(context, builder, sig, args):
    tpe, = sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    call(context, builder, libawkward.FillableArray_endrecord, (proxyin.rawptr,))
    return context.get_dummy_value()
